"""
    :codeauthor: Rupesh Tare <rupesht@saltstack.com>
"""


import logging

import pytest

import salt.modules.pw_user as pw_user
from salt.exceptions import CommandExecutionError
from tests.support.mock import MagicMock, patch

pwd = pytest.importorskip(
    "pwd", reason="Tests can only run on systems with the python pwd module"
)

log = logging.getLogger(__name__)


@pytest.fixture
def configure_loader_modules():
    return {pw_user: {}}


def test_add():
    """
    Test for adding a user
    """
    with patch.dict(pw_user.__grains__, {"os_family": "RedHat"}):
        mock = MagicMock(return_value=0)
        with patch.dict(pw_user.__salt__, {"cmd.retcode": mock}):
            assert pw_user.add("a")


def test_delete():
    """
    Test for deleting a user
    """
    mock = MagicMock(return_value=0)
    with patch.dict(pw_user.__salt__, {"cmd.retcode": mock}):
        assert pw_user.delete("A")


def test_getent():
    """
    Test if user.getent already have a value
    """
    mock_user = "saltdude"

    class MockData:
        pw_name = mock_user

    with patch("pwd.getpwall", MagicMock(return_value=[MockData()])):
        with patch.dict(pw_user.__context__, {"user.getent": mock_user}):
            assert pw_user.getent() == mock_user

            with patch.object(pw_user, "info", MagicMock(return_value=mock_user)):
                assert pw_user.getent(True)[0] == mock_user


def test_chuid():
    """
    Test if user id given is same as previous id
    """
    mock = MagicMock(return_value={"uid": "A"})
    with patch.object(pw_user, "info", mock):
        assert pw_user.chuid("name", "A")

    mock = MagicMock(return_value=None)
    with patch.dict(pw_user.__salt__, {"cmd.run": mock}):
        mock = MagicMock(side_effect=[{"uid": "A"}, {"uid": "A"}])
        with patch.object(pw_user, "info", mock):
            assert not pw_user.chuid("name", "B")

    mock = MagicMock(return_value=None)
    with patch.dict(pw_user.__salt__, {"cmd.run": mock}):
        mock = MagicMock(side_effect=[{"uid": "A"}, {"uid": "B"}])
        with patch.object(pw_user, "info", mock):
            assert pw_user.chuid("name", "A")


def test_chgid():
    """
    Test if group id given is same as previous id
    """
    mock = MagicMock(return_value={"gid": 1})
    with patch.object(pw_user, "info", mock):
        assert pw_user.chgid("name", 1)

    mock = MagicMock(return_value=None)
    with patch.dict(pw_user.__salt__, {"cmd.run": mock}):
        mock = MagicMock(side_effect=[{"gid": 2}, {"gid": 2}])
        with patch.object(pw_user, "info", mock):
            assert not pw_user.chgid("name", 1)

    mock = MagicMock(return_value=None)
    with patch.dict(pw_user.__salt__, {"cmd.run": mock}):
        mock = MagicMock(side_effect=[{"gid": 1}, {"gid": 2}])
        with patch.object(pw_user, "info", mock):
            assert pw_user.chgid("name", 1)


def test_chshell():
    """
    Test if shell given is same as previous shell
    """
    mock = MagicMock(return_value={"shell": "A"})
    with patch.object(pw_user, "info", mock):
        assert pw_user.chshell("name", "A")

    mock = MagicMock(return_value=None)
    with patch.dict(pw_user.__salt__, {"cmd.run": mock}):
        mock = MagicMock(side_effect=[{"shell": "B"}, {"shell": "B"}])
        with patch.object(pw_user, "info", mock):
            assert not pw_user.chshell("name", "A")

    mock = MagicMock(return_value=None)
    with patch.dict(pw_user.__salt__, {"cmd.run": mock}):
        mock = MagicMock(side_effect=[{"shell": "A"}, {"shell": "B"}])
        with patch.object(pw_user, "info", mock):
            assert pw_user.chshell("name", "A")


def test_chhome():
    """
    Test if home directory given is same as previous home directory
    """
    mock = MagicMock(return_value={"home": "A"})
    with patch.object(pw_user, "info", mock):
        assert pw_user.chhome("name", "A")

    mock = MagicMock(return_value=None)
    with patch.dict(pw_user.__salt__, {"cmd.run": mock}):
        mock = MagicMock(side_effect=[{"home": "B"}, {"home": "B"}])
        with patch.object(pw_user, "info", mock):
            assert not pw_user.chhome("name", "A")

    mock = MagicMock(return_value=None)
    with patch.dict(pw_user.__salt__, {"cmd.run": mock}):
        mock = MagicMock(side_effect=[{"home": "A"}, {"home": "B"}])
        with patch.object(pw_user, "info", mock):
            assert pw_user.chhome("name", "A")


def test_chgroups():
    """
    Test if no group needs to be added
    """
    mock = MagicMock(return_value=False)
    with patch.dict(pw_user.__salt__, {"cmd.retcode": mock}):
        mock = MagicMock(return_value=["a", "b", "c", "d"])
        with patch.object(pw_user, "list_groups", mock):
            assert pw_user.chgroups("name", "a, b, c, d")

    mock = MagicMock(return_value=False)
    with patch.dict(pw_user.__salt__, {"cmd.retcode": mock}):
        mock = MagicMock(return_value=["a", "b"])
        with patch.object(pw_user, "list_groups", mock):
            assert pw_user.chgroups("name", "a, b, c")


def test_chfullname():
    """
    Change the user's Full Name
    """
    mock = MagicMock(return_value=False)
    with patch.object(pw_user, "_get_gecos", mock):
        assert not pw_user.chfullname("name", "fullname")

    mock = MagicMock(return_value={"fullname": "fullname"})
    with patch.object(pw_user, "_get_gecos", mock):
        assert pw_user.chfullname("name", "fullname")

    mock = MagicMock(return_value={"fullname": "Unicøde name ①③②"})
    with patch.object(pw_user, "_get_gecos", mock):
        assert pw_user.chfullname("name", "Unicøde name ①③②")

    mock = MagicMock(return_value={"fullname": "fullname"})
    with patch.object(pw_user, "_get_gecos", mock):
        mock = MagicMock(return_value=None)
        with patch.dict(pw_user.__salt__, {"cmd.run": mock}):
            mock = MagicMock(return_value={"fullname": "fullname2"})
            with patch.object(pw_user, "info", mock):
                assert not pw_user.chfullname("name", "fullname1")

    mock = MagicMock(return_value={"fullname": "fullname2"})
    with patch.object(pw_user, "_get_gecos", mock):
        mock = MagicMock(return_value=None)
        with patch.dict(pw_user.__salt__, {"cmd.run": mock}):
            mock = MagicMock(return_value={"fullname": "fullname2"})
            with patch.object(pw_user, "info", mock):
                assert not pw_user.chfullname("name", "fullname1")


def test_chroomnumber():
    """
    Change the user's Room Number
    """
    mock = MagicMock(return_value=False)
    with patch.object(pw_user, "_get_gecos", mock):
        assert not pw_user.chroomnumber("name", 1)

    mock = MagicMock(return_value={"roomnumber": "Unicøde room ①③②"})
    with patch.object(pw_user, "_get_gecos", mock):
        assert pw_user.chroomnumber("name", "Unicøde room ①③②")

    mock = MagicMock(return_value={"roomnumber": "1"})
    with patch.object(pw_user, "_get_gecos", mock):
        assert pw_user.chroomnumber("name", 1)

    mock = MagicMock(return_value={"roomnumber": "2"})
    with patch.object(pw_user, "_get_gecos", mock):
        mock = MagicMock(return_value=None)
        with patch.dict(pw_user.__salt__, {"cmd.run": mock}):
            mock = MagicMock(return_value={"roomnumber": "3"})
            with patch.object(pw_user, "info", mock):
                assert not pw_user.chroomnumber("name", 1)

    mock = MagicMock(return_value={"roomnumber": "3"})
    with patch.object(pw_user, "_get_gecos", mock):
        mock = MagicMock(return_value=None)
        with patch.dict(pw_user.__salt__, {"cmd.run": mock}):
            mock = MagicMock(return_value={"roomnumber": "3"})
            with patch.object(pw_user, "info", mock):
                assert not pw_user.chroomnumber("name", 1)


def test_chworkphone():
    """
    Change the user's Work Phone
    """
    mock = MagicMock(return_value=False)
    with patch.object(pw_user, "_get_gecos", mock):
        assert not pw_user.chworkphone("name", 1)

    mock = MagicMock(return_value={"workphone": "1"})
    with patch.object(pw_user, "_get_gecos", mock):
        assert pw_user.chworkphone("name", 1)

    mock = MagicMock(return_value={"workphone": "Unicøde phone number ①③②"})
    with patch.object(pw_user, "_get_gecos", mock):
        assert pw_user.chworkphone("name", "Unicøde phone number ①③②")

    mock = MagicMock(return_value={"workphone": "2"})
    with patch.object(pw_user, "_get_gecos", mock):
        mock = MagicMock(return_value=None)
        with patch.dict(pw_user.__salt__, {"cmd.run": mock}):
            mock = MagicMock(return_value={"workphone": "3"})
            with patch.object(pw_user, "info", mock):
                assert not pw_user.chworkphone("name", 1)

    mock = MagicMock(return_value={"workphone": "3"})
    with patch.object(pw_user, "_get_gecos", mock):
        mock = MagicMock(return_value=None)
        with patch.dict(pw_user.__salt__, {"cmd.run": mock}):
            mock = MagicMock(return_value={"workphone": "3"})
            with patch.object(pw_user, "info", mock):
                assert not pw_user.chworkphone("name", 1)


def test_chhomephone():
    """
    Change the user's Home Phone
    """
    mock = MagicMock(return_value=False)
    with patch.object(pw_user, "_get_gecos", mock):
        assert not pw_user.chhomephone("name", 1)

    mock = MagicMock(return_value={"homephone": "1"})
    with patch.object(pw_user, "_get_gecos", mock):
        assert pw_user.chhomephone("name", 1)

    mock = MagicMock(return_value={"homephone": "Unicøde phone number ①③②"})
    with patch.object(pw_user, "_get_gecos", mock):
        assert pw_user.chhomephone("name", "Unicøde phone number ①③②")

    mock = MagicMock(return_value={"homephone": "2"})
    with patch.object(pw_user, "_get_gecos", mock):
        mock = MagicMock(return_value=None)
        with patch.dict(pw_user.__salt__, {"cmd.run": mock}):
            mock = MagicMock(return_value={"homephone": "3"})
            with patch.object(pw_user, "info", mock):
                assert not pw_user.chhomephone("name", 1)

    mock = MagicMock(return_value={"homephone": "3"})
    with patch.object(pw_user, "_get_gecos", mock):
        mock = MagicMock(return_value=None)
        with patch.dict(pw_user.__salt__, {"cmd.run": mock}):
            mock = MagicMock(return_value={"homephone": "3"})
            with patch.object(pw_user, "info", mock):
                assert not pw_user.chhomephone("name", 1)


def test_info():
    """
    Return user information
    """
    assert pw_user.info("name") == {}

    mock = MagicMock(
        return_value=pwd.struct_passwd(
            (
                "_TEST_GROUP",
                "*",
                83,
                83,
                "AMaViS Daemon",
                "/var/virusmails",
                "/usr/bin/false",
            )
        )
    )
    with patch.object(pwd, "getpwnam", mock):
        mock = MagicMock(return_value="Group Name")
        with patch.object(pw_user, "list_groups", mock):
            assert pw_user.info("name")["name"] == "_TEST_GROUP"


def test_list_groups():
    """
    Return a list of groups the named user belongs to
    """
    mock_group = "saltgroup"

    with patch("salt.utils.user.get_group_list", MagicMock(return_value=[mock_group])):
        assert pw_user.list_groups("name") == [mock_group]


def test_list_users():
    """
    Return a list of all users
    """
    mock_user = "saltdude"

    class MockData:
        pw_name = mock_user

    with patch("pwd.getpwall", MagicMock(return_value=[MockData()])):
        assert pw_user.list_users() == [mock_user]


def test_rename():
    """
    Change the username for a named user
    """
    mock = MagicMock(return_value=False)
    with patch.object(pw_user, "info", mock):
        pytest.raises(CommandExecutionError, pw_user.rename, "name", 1)

    mock = MagicMock(return_value=True)
    with patch.object(pw_user, "info", mock):
        pytest.raises(CommandExecutionError, pw_user.rename, "name", 1)

    mock = MagicMock(return_value=None)
    with patch.dict(pw_user.__salt__, {"cmd.run": mock}):
        mock = MagicMock(side_effect=[{"name": ""}, False, {"name": "name"}])
        with patch.object(pw_user, "info", mock):
            assert pw_user.rename("name", "name")

    mock = MagicMock(return_value=None)
    with patch.dict(pw_user.__salt__, {"cmd.run": mock}):
        mock = MagicMock(side_effect=[{"name": ""}, False, {"name": ""}])
        with patch.object(pw_user, "info", mock):
            assert not pw_user.rename("name", "name")



"""
Management of Influxdb databases
================================

(compatible with InfluxDB version 0.9+)
"""


def __virtual__():
    """
    Only load if the influxdb module is available
    """
    if "influxdb.db_exists" in __salt__:
        return "influxdb_database"
    return (False, "influxdb module could not be loaded")


def present(name, **client_args):
    """
    Ensure that given database is present.

    name
        Name of the database to create.
    """
    ret = {
        "name": name,
        "changes": {},
        "result": True,
        "comment": "Database {} is already present".format(name),
    }

    if not __salt__["influxdb.db_exists"](name, **client_args):
        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = "Database {} is absent and will be created".format(name)
            return ret
        if __salt__["influxdb.create_db"](name, **client_args):
            ret["comment"] = "Database {} has been created".format(name)
            ret["changes"][name] = "Present"
            return ret
        else:
            ret["comment"] = "Failed to create database {}".format(name)
            ret["result"] = False
            return ret

    return ret


def absent(name, **client_args):
    """
    Ensure that given database is absent.

    name
        Name of the database to remove.
    """
    ret = {
        "name": name,
        "changes": {},
        "result": True,
        "comment": "Database {} is not present".format(name),
    }

    if __salt__["influxdb.db_exists"](name, **client_args):
        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = "Database {} is present and needs to be removed".format(
                name
            )
            return ret
        if __salt__["influxdb.drop_db"](name, **client_args):
            ret["comment"] = "Database {} has been removed".format(name)
            ret["changes"][name] = "Absent"
            return ret
        else:
            ret["comment"] = "Failed to remove database {}".format(name)
            ret["result"] = False
            return ret

    return ret



import logging
import shutil

import pytest

import salt.cache
import salt.loader
from tests.pytests.functional.cache.helpers import run_common_cache_tests

log = logging.getLogger(__name__)


@pytest.fixture
def cache(minion_opts):
    opts = minion_opts.copy()
    opts["cache"] = "localfs"
    cache = salt.cache.factory(opts)
    try:
        yield cache
    finally:
        shutil.rmtree(opts["cachedir"], ignore_errors=True)


def test_caching(subtests, cache):
    run_common_cache_tests(subtests, cache)



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""


import types

import pytest

import salt.modules.win_dns_client as win_dns_client
import salt.utils.stringutils
from tests.support.mock import MagicMock, Mock, patch

wmi = pytest.importorskip("wmi", reason="WMI only available on Windows")


class Mockwmi:
    """
    Mock wmi class
    """

    def __init__(self):
        pass

    NetConnectionID = "Local Area Connection"
    Index = 0
    DNSServerSearchOrder = ["10.1.1.10"]
    Description = "Local Area Connection"
    DHCPEnabled = True


class Mockwinapi:
    """
    Mock winapi class
    """

    def __init__(self):
        pass

    class winapi:
        """
        Mock winapi class
        """

        def __init__(self):
            pass

        @staticmethod
        def Com():
            """
            Mock Com method
            """
            return True


@pytest.fixture
def configure_loader_modules():
    mock_pythoncom = types.ModuleType(salt.utils.stringutils.to_str("pythoncom"))
    with patch.dict("sys.modules", {"pythoncom": mock_pythoncom}):
        yield {win_dns_client: {"wmi": wmi}}


def test_get_dns_servers():
    """
    Test if it return a list of the configured DNS servers
    of the specified interface.
    """
    WMI = Mock()
    with patch("salt.utils.winapi.Com", MagicMock()), patch.object(
        WMI, "Win32_NetworkAdapter", return_value=[Mockwmi()]
    ), patch.object(
        WMI, "Win32_NetworkAdapterConfiguration", return_value=[Mockwmi()]
    ), patch.object(
        wmi, "WMI", Mock(return_value=WMI)
    ):
        assert win_dns_client.get_dns_servers("Local Area Connection") == ["10.1.1.10"]

        assert not win_dns_client.get_dns_servers("Ethernet")


def test_rm_dns():
    """
    Test if it remove the DNS server from the network interface.
    """
    with patch.dict(
        win_dns_client.__salt__, {"cmd.retcode": MagicMock(return_value=0)}
    ):
        assert win_dns_client.rm_dns("10.1.1.10")


def test_add_dns():
    """
    Test if it add the DNS server to the network interface.
    """
    WMI = Mock()
    with patch("salt.utils.winapi.Com", MagicMock()), patch.object(
        WMI, "Win32_NetworkAdapter", return_value=[Mockwmi()]
    ), patch.object(
        WMI, "Win32_NetworkAdapterConfiguration", return_value=[Mockwmi()]
    ), patch.object(
        wmi, "WMI", Mock(return_value=WMI)
    ):
        assert not win_dns_client.add_dns("10.1.1.10", "Ethernet")

        assert win_dns_client.add_dns("10.1.1.10", "Local Area Connection")

    with patch.object(
        win_dns_client, "get_dns_servers", MagicMock(return_value=["10.1.1.10"])
    ), patch.dict(
        win_dns_client.__salt__, {"cmd.retcode": MagicMock(return_value=0)}
    ), patch.object(
        wmi, "WMI", Mock(return_value=WMI)
    ):
        assert win_dns_client.add_dns("10.1.1.0", "Local Area Connection")


def test_dns_dhcp():
    """
    Test if it configure the interface to get its
    DNS servers from the DHCP server
    """
    with patch.dict(
        win_dns_client.__salt__, {"cmd.retcode": MagicMock(return_value=0)}
    ):
        assert win_dns_client.dns_dhcp()


def test_get_dns_config():
    """
    Test if it get the type of DNS configuration (dhcp / static)
    """
    WMI = Mock()
    with patch("salt.utils.winapi.Com", MagicMock()), patch.object(
        WMI, "Win32_NetworkAdapter", return_value=[Mockwmi()]
    ), patch.object(
        WMI, "Win32_NetworkAdapterConfiguration", return_value=[Mockwmi()]
    ), patch.object(
        wmi, "WMI", Mock(return_value=WMI)
    ):
        assert win_dns_client.get_dns_config()


def test___virtual__non_windows():
    mock_false = Mock(return_value=False)
    with patch("salt.utils.platform.is_windows", mock_false):
        result = win_dns_client.__virtual__()
        expected = (
            False,
            "Module win_dns_client: module only works on Windows systems",
        )
        assert result == expected


def test___virtual__missing_libs():
    with patch.object(win_dns_client, "HAS_LIBS", False):
        result = win_dns_client.__virtual__()
        expected = (False, "Module win_dns_client: missing required libraries")
        assert result == expected


def test___virtual__():
    result = win_dns_client.__virtual__()
    expected = "win_dns_client"
    assert result == expected



"""
A simple beacon to watch journald for specific entries
"""
import logging

import salt.utils.beacons
import salt.utils.data

try:
    import systemd.journal  # pylint: disable=no-name-in-module

    HAS_SYSTEMD = True
except ImportError:
    HAS_SYSTEMD = False


log = logging.getLogger(__name__)

__virtualname__ = "journald"


def __virtual__():
    if HAS_SYSTEMD:
        return __virtualname__
    err_msg = "systemd library is missing."
    log.error("Unable to load %s beacon: %s", __virtualname__, err_msg)
    return False, err_msg


def _get_journal():
    """
    Return the active running journal object
    """
    if "systemd.journald" in __context__:
        return __context__["systemd.journald"]
    __context__["systemd.journald"] = systemd.journal.Reader()
    # get to the end of the journal
    __context__["systemd.journald"].seek_tail()
    __context__["systemd.journald"].get_previous()
    return __context__["systemd.journald"]


def validate(config):
    """
    Validate the beacon configuration
    """
    # Configuration for journald beacon should be a list of dicts
    if not isinstance(config, list):
        return (False, "Configuration for journald beacon must be a list.")
    else:
        config = salt.utils.beacons.list_to_dict(config)

        for name in config.get("services", {}):
            if not isinstance(config["services"][name], dict):
                return (
                    False,
                    "Services configuration for journald beacon must be a list of"
                    " dictionaries.",
                )
    return True, "Valid beacon configuration"


def beacon(config):
    """
    The journald beacon allows for the systemd journal to be parsed and linked
    objects to be turned into events.

    This beacons config will return all sshd jornal entries

    .. code-block:: yaml

        beacons:
          journald:
            - services:
                sshd:
                  SYSLOG_IDENTIFIER: sshd
                  PRIORITY: 6
    """
    ret = []
    journal = _get_journal()

    config = salt.utils.beacons.list_to_dict(config)

    while True:
        cur = journal.get_next()
        if not cur:
            break

        for name in config.get("services", {}):
            n_flag = 0
            for key in config["services"][name]:
                if isinstance(key, str):
                    key = salt.utils.data.decode(key)
                if key in cur:
                    if config["services"][name][key] == cur[key]:
                        n_flag += 1
            if n_flag == len(config["services"][name]):
                # Match!
                sub = salt.utils.data.simple_types_filter(cur)
                sub.update({"tag": name})
                ret.append(sub)
    return ret



import logging
import os

import pytest

import salt.utils.files
from salt.beacons import inotify

try:
    import pyinotify  # pylint: disable=unused-import

    HAS_PYINOTIFY = True
except ImportError:
    HAS_PYINOTIFY = False


log = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.skipif(HAS_PYINOTIFY is False, reason="pyinotify is not available."),
]


@pytest.fixture
def configure_loader_modules():
    return {inotify: {}}


@pytest.fixture(autouse=True)
def _close_inotify(configure_loader_modules):
    try:
        yield
    finally:
        inotify.close({})


def test_non_list_config():
    config = {}

    ret = inotify.validate(config)

    assert ret == (False, "Configuration for inotify beacon must be a list.")


def test_empty_config():
    config = [{}]
    ret = inotify.validate(config)
    _expected = (False, "Configuration for inotify beacon must include files.")
    assert ret == _expected


def test_files_none_config():
    config = [{"files": None}]
    ret = inotify.validate(config)
    _expected = (
        False,
        "Configuration for inotify beacon invalid, files must be a dict.",
    )
    assert ret == _expected


def test_files_list_config():
    config = [{"files": [{"/importantfile": {"mask": ["modify"]}}]}]
    ret = inotify.validate(config)
    _expected = (
        False,
        "Configuration for inotify beacon invalid, files must be a dict.",
    )
    assert ret == _expected


@pytest.mark.skip_on_freebsd(
    reason="Skip on FreeBSD - does not yet have full inotify/watchdog support",
)
def test_file_open():
    path = os.path.realpath(__file__)
    config = [{"files": {path: {"mask": ["open"]}}}]
    ret = inotify.validate(config)
    assert ret == (True, "Valid beacon configuration")

    ret = inotify.beacon(config)
    assert ret == []

    with salt.utils.files.fopen(path, "r") as f:
        pass
    ret = inotify.beacon(config)
    assert len(ret) == 1
    assert ret[0]["path"] == path
    assert ret[0]["change"] == "IN_OPEN"


@pytest.mark.skip_on_freebsd(
    reason="Skip on FreeBSD - does not yet have full inotify/watchdog support",
)
def test_dir_no_auto_add(tmp_path):
    config = [{"files": {str(tmp_path): {"mask": ["create"]}}}]
    ret = inotify.validate(config)
    assert ret == (True, "Valid beacon configuration")

    ret = inotify.beacon(config)
    assert ret == []
    fp = str(tmp_path / "tmpfile")
    with salt.utils.files.fopen(fp, "w") as f:
        pass
    ret = inotify.beacon(config)
    assert len(ret) == 1
    assert ret[0]["path"] == fp
    assert ret[0]["change"] == "IN_CREATE"
    with salt.utils.files.fopen(fp, "r") as f:
        pass
    ret = inotify.beacon(config)
    assert ret == []


@pytest.mark.skip_on_freebsd(
    reason="Skip on FreeBSD - does not yet have full inotify/watchdog support",
)
def test_dir_auto_add(tmp_path):
    config = [
        {"files": {str(tmp_path): {"mask": ["create", "open"], "auto_add": True}}}
    ]
    ret = inotify.validate(config)
    assert ret == (True, "Valid beacon configuration")

    ret = inotify.beacon(config)
    assert ret == []
    fp = str(tmp_path / "tmpfile")
    with salt.utils.files.fopen(fp, "w") as f:
        pass
    ret = inotify.beacon(config)
    assert len(ret) == 2
    assert ret[0]["path"] == fp
    assert ret[0]["change"] == "IN_CREATE"
    assert ret[1]["path"] == fp
    assert ret[1]["change"] == "IN_OPEN"
    with salt.utils.files.fopen(fp, "r") as f:
        pass
    ret = inotify.beacon(config)
    assert len(ret) == 1
    assert ret[0]["path"] == fp
    assert ret[0]["change"] == "IN_OPEN"


@pytest.mark.skip_on_freebsd(
    reason="Skip on FreeBSD - does not yet have full inotify/watchdog support",
)
def test_dir_recurse(tmp_path):
    dp1 = str(tmp_path / "subdir1")
    os.mkdir(dp1)
    dp2 = os.path.join(dp1, "subdir2")
    os.mkdir(dp2)
    fp = os.path.join(dp2, "tmpfile")
    with salt.utils.files.fopen(fp, "w") as f:
        pass
    config = [{"files": {str(tmp_path): {"mask": ["open"], "recurse": True}}}]
    ret = inotify.validate(config)
    assert ret == (True, "Valid beacon configuration")

    ret = inotify.beacon(config)
    assert ret == []
    with salt.utils.files.fopen(fp) as f:
        pass
    ret = inotify.beacon(config)
    assert len(ret) == 3
    assert ret[0]["path"] == dp1
    assert ret[0]["change"] == "IN_OPEN|IN_ISDIR"
    assert ret[1]["path"] == dp2
    assert ret[1]["change"] == "IN_OPEN|IN_ISDIR"
    assert ret[2]["path"] == fp
    assert ret[2]["change"] == "IN_OPEN"


@pytest.mark.skip_on_freebsd(
    reason="Skip on FreeBSD - does not yet have full inotify/watchdog support",
)
def test_dir_recurse_auto_add(tmp_path):
    dp1 = str(tmp_path / "subdir1")
    os.mkdir(dp1)
    config = [
        {
            "files": {
                str(tmp_path): {
                    "mask": ["create", "delete"],
                    "recurse": True,
                    "auto_add": True,
                }
            }
        }
    ]
    ret = inotify.validate(config)
    assert ret == (True, "Valid beacon configuration")

    ret = inotify.beacon(config)
    assert ret == []
    dp2 = os.path.join(dp1, "subdir2")
    os.mkdir(dp2)
    ret = inotify.beacon(config)
    assert len(ret) == 1
    assert ret[0]["path"] == dp2
    assert ret[0]["change"] == "IN_CREATE|IN_ISDIR"
    fp = os.path.join(dp2, "tmpfile")
    with salt.utils.files.fopen(fp, "w") as f:
        pass
    ret = inotify.beacon(config)
    assert len(ret) == 1
    assert ret[0]["path"] == fp
    assert ret[0]["change"] == "IN_CREATE"
    os.remove(fp)
    ret = inotify.beacon(config)
    assert len(ret) == 1
    assert ret[0]["path"] == fp
    assert ret[0]["change"] == "IN_DELETE"


@pytest.mark.skip_on_freebsd(
    reason="Skip on FreeBSD - does not yet have full inotify/watchdog support",
)
def test_multi_files_exclude(tmp_path):
    dp1 = str(tmp_path / "subdir1")
    dp2 = str(tmp_path / "subdir2")
    os.mkdir(dp1)
    os.mkdir(dp2)
    _exclude1 = "{}/subdir1/*tmpfile*$".format(str(tmp_path))
    _exclude2 = "{}/subdir2/*filetmp*$".format(str(tmp_path))
    config = [
        {
            "files": {
                dp1: {
                    "mask": ["create", "delete"],
                    "recurse": True,
                    "exclude": [{_exclude1: {"regex": True}}],
                    "auto_add": True,
                }
            }
        },
        {
            "files": {
                dp2: {
                    "mask": ["create", "delete"],
                    "recurse": True,
                    "exclude": [{_exclude2: {"regex": True}}],
                    "auto_add": True,
                }
            }
        },
    ]
    ret = inotify.validate(config)
    assert ret == (True, "Valid beacon configuration")

    fp = os.path.join(dp1, "tmpfile")
    with salt.utils.files.fopen(fp, "w") as f:
        pass
    ret = inotify.beacon(config)
    assert len(ret) == 0
    os.remove(fp)
    ret = inotify.beacon(config)
    assert len(ret) == 0

    fp = os.path.join(dp2, "tmpfile")
    with salt.utils.files.fopen(fp, "w") as f:
        pass
    ret = inotify.beacon(config)
    assert len(ret) == 1
    assert ret[0]["path"] == fp
    assert ret[0]["change"] == "IN_CREATE"
    os.remove(fp)
    ret = inotify.beacon(config)
    assert len(ret) == 1
    assert ret[0]["path"] == fp
    assert ret[0]["change"] == "IN_DELETE"


# Check __get_notifier and ensure that the right bits are in __context__
# including a beacon_name specific notifier is found.
def test__get_notifier():
    config = {
        "files": {
            "/tmp/httpd/vhost.d": {
                "mask": ["delete", "modify"],
                "recurse": True,
                "auto_add": True,
                "exclude": [
                    {"/tmp/httpd/vhost.d/.+?\\.sw[px]*$|4913|~$": {"regex": True}}
                ],
            },
            "/tmp/httpd/conf.d": {
                "mask": ["delete", "modify"],
                "recurse": True,
                "auto_add": True,
                "exclude": [
                    {"/tmp/httpd/vhost.d/.+?\\.sw[px]*$|4913|~$": {"regex": True}}
                ],
            },
            "/tmp/httpd/conf": {
                "mask": ["delete", "modify"],
                "recurse": True,
                "auto_add": True,
                "exclude": [
                    {"/tmp/httpd/vhost.d/.+?\\.sw[px]*$|4913|~$": {"regex": True}}
                ],
            },
        },
        "coalesce": True,
        "beacon_module": "inotify",
        "_beacon_name": "httpd.inotify",
    }

    ret = inotify._get_notifier(config)
    assert "inotify.queue" in inotify.__context__
    assert "httpd.inotify.notifier" in inotify.__context__



"""
    :codeauthor: Jorge Schrauwen <info@blackdot.be>

    TestCase for salt.modules.smartos_imgadm module
"""


import pytest

import salt.modules.smartos_imgadm as imgadm
from salt.modules.smartos_imgadm import _parse_image_meta


@pytest.fixture
def image_orphan():
    return {
        "manifest": {"uuid": "07f360fd-12d5-e624-a279-eb8a15b630f6"},
        "zpool": "zones",
        "cloneNames": [],
        "clones": 0,
    }


@pytest.fixture
def image_native():
    return {
        "manifest": {
            "v": 2,
            "uuid": "9d91e334-3bcf-11e8-bb0b-e7b49eb01e38",
            "owner": "00000000-0000-0000-0000-000000000000",
            "name": "pkgbuild",
            "version": "18.1.0",
            "state": "active",
            "disabled": False,
            "public": True,
            "published_at": "2018-04-09T08:25:52Z",
            "type": "zone-dataset",
            "os": "smartos",
            "files": [
                {
                    "sha1": "5efaf95b7f226eb09c7d5e6c3734f8aa654b811d",
                    "size": 465411979,
                    "compression": "gzip",
                }
            ],
            "description": "A SmartOS image pre-configured for building pkgsrc packages.",
            "homepage": "https://docs.joyent.com/images/smartos/pkgbuild",
            "urn": "sdc:sdc:pkgbuild:18.1.0",
            "requirements": {
                "min_platform": {"7.0": "20141030T081701Z"},
                "networks": [{"name": "net0", "description": "public"}],
            },
            "tags": {"role": "os", "group": "pkgbuild"},
        },
        "zpool": "zones",
        "source": "https://images.joyent.com",
        "cloneNames": ["zones/dda70f61-70fe-65e7-cf70-d878d69442d4"],
        "clones": 1,
    }


@pytest.fixture
def image_lx():
    return {
        "manifest": {
            "v": 2,
            "uuid": "05140a7e-279f-11e6-aedf-47d4f69d2887",
            "owner": "00000000-0000-0000-0000-000000000000",
            "name": "ubuntu-16.04",
            "version": "20160601",
            "state": "active",
            "disabled": False,
            "public": True,
            "published_at": "2016-06-01T02:17:41Z",
            "type": "lx-dataset",
            "os": "linux",
            "files": [
                {
                    "sha1": "d342f137c5ccef0702ec479acb63c196cf81b38a",
                    "size": 134969110,
                    "compression": "gzip",
                }
            ],
            "description": (
                "Container-native Ubuntu 16.04 64-bit image. Built to run on containers"
                " with bare metal speed, while offering all the services of a typical unix"
                " host."
            ),
            "homepage": "https://docs.joyent.com/images/container-native-linux",
            "requirements": {
                "networks": [{"name": "net0", "description": "public"}],
                "min_platform": {"7.0": "20160225T122859Z"},
                "brand": "lx",
            },
            "tags": {"role": "os", "kernel_version": "4.3.0"},
        },
        "zpool": "zones",
        "source": "https://images.joyent.com",
        "cloneNames": ["zones/e4c1f6b5-4429-e6c2-ae2a-d6aa58bdeebb"],
        "clones": 1,
    }


@pytest.fixture
def image_zvol():
    return {
        "manifest": {
            "v": 2,
            "uuid": "ac99517a-72ac-44c0-90e6-c7ce3d944a0a",
            "owner": "00000000-0000-0000-0000-000000000000",
            "name": "ubuntu-certified-18.04",
            "version": "20180808",
            "state": "active",
            "disabled": False,
            "public": True,
            "published_at": "2018-10-11T12:45:24.804Z",
            "type": "zvol",
            "os": "linux",
            "files": [
                {
                    "sha1": "9f7704969507bd97e160a8f42a3631487644e457",
                    "size": 372276887,
                    "compression": "gzip",
                }
            ],
            "description": (
                "Ubuntu 18.04 LTS (20180808 64-bit). Certified Ubuntu Server Cloud Image"
                " from Canonical. For kvm and bhyve."
            ),
            "homepage": "https://docs.joyent.com/images/linux/ubuntu-certified",
            "requirements": {
                "min_platform": {"7.0": "20150929T232348Z"},
                "networks": [{"name": "net0", "description": "public"}],
                "ssh_key": True,
            },
            "nic_driver": "virtio",
            "disk_driver": "virtio",
            "cpu_type": "host",
            "image_size": 10240,
            "tags": {"default_user": "ubuntu", "role": "os"},
        },
        "zpool": "zones",
        "source": "https://images.joyent.com",
        "cloneNames": [],
        "clones": 0,
    }


@pytest.fixture
def image_docker():
    return {
        "manifest": {
            "v": 2,
            "uuid": "4a3db8cb-0e94-ae23-588c-ee7934088927",
            "owner": "00000000-0000-0000-0000-000000000000",
            "name": "docker-layer",
            "version": "62487cf6a7f6",
            "disabled": False,
            "public": True,
            "published_at": "2019-03-23T01:32:25.320Z",
            "type": "docker",
            "os": "linux",
            "description": '/bin/sh -c #(nop)  CMD ["/bin/bash" "/opt/start.sh" "-bash"]',
            "tags": {
                "docker:repo": "busybox42/zimbra-docker-centos",
                "docker:id": "sha256:62487cf6a7f698af4edc20707e14b1b3bba13b98bea3375f05af04859a30b222",
                "docker:architecture": "amd64",
                "docker:tag:latest": True,
                "docker:config": {
                    "Cmd": ["/bin/bash", "/opt/start.sh", "-bash"],
                    "Entrypoint": None,
                    "Env": [
                        "PATH=/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin"
                    ],
                    "WorkingDir": "",
                },
            },
            "origin": "2f0c529b-7bab-28d1-ff34-bdc9281b7a4b",
        },
        "zpool": "zones",
        "source": "https://docker.io",
        "cloneNames": [],
        "clones": 0,
    }


@pytest.fixture
def configure_loader_modules():
    return {imgadm: {}}


def test_parse_image_meta_orphan(image_orphan):
    """
    Test the internal _parse_image_meta methode
    Feed it an 'orphan' image as we get it from from imgadm list -j
    """
    ret = {"Error": "This looks like an orphaned image, image payload was invalid."}
    assert _parse_image_meta(image_orphan, True) == ret


def test_parse_image_meta_native(image_native):
    """
    Test the internal _parse_image_meta methode
    Feed it an 'native' image as we get it from from imgadm list -j
    """
    ret = {
        "description": ("A SmartOS image pre-configured for building pkgsrc packages."),
        "name": "pkgbuild",
        "os": "smartos",
        "published": "2018-04-09T08:25:52Z",
        "source": "https://images.joyent.com",
        "version": "18.1.0",
    }
    assert _parse_image_meta(image_native, True) == ret


def test_parse_image_meta_lx(image_lx):
    """
    Test the internal _parse_image_meta methode
    Feed it an 'lx' image as we get it from from imgadm list -j
    """
    ret = {
        "description": (
            "Container-native Ubuntu 16.04 64-bit image. Built to run on "
            "containers with bare metal speed, while offering all the "
            "services of a typical unix host."
        ),
        "name": "ubuntu-16.04",
        "os": "linux",
        "published": "2016-06-01T02:17:41Z",
        "source": "https://images.joyent.com",
        "version": "20160601",
    }
    assert _parse_image_meta(image_lx, True) == ret


def test_parse_image_meta_zvol(image_zvol):
    """
    Test the internal _parse_image_meta methode
    Feed it an 'zvol' image as we get it from from imgadm list -j
    """
    ret = {
        "description": (
            "Ubuntu 18.04 LTS (20180808 64-bit). Certified Ubuntu Server "
            "Cloud Image from Canonical. For kvm and bhyve."
        ),
        "name": "ubuntu-certified-18.04",
        "os": "linux",
        "published": "2018-10-11T12:45:24.804Z",
        "source": "https://images.joyent.com",
        "version": "20180808",
    }
    assert _parse_image_meta(image_zvol, True) == ret


def test_parse_image_meta_docker(image_docker):
    """
    Test the internal _parse_image_meta methode
    Feed it an 'docker' image as we get it from from imgadm list -j
    """
    ret = {
        "description": (
            "Docker image imported from "
            "busybox42/zimbra-docker-centos:latest on "
            "2019-03-23T01:32:25.320Z."
        ),
        "name": "busybox42/zimbra-docker-centos:latest",
        "os": "linux",
        "published": "2019-03-23T01:32:25.320Z",
        "source": "https://docker.io",
        "version": "62487cf6a7f6",
    }
    assert _parse_image_meta(image_docker, True) == ret



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>

    Test cases for salt.modules.pecl
"""


import pytest

import salt.modules.pecl as pecl
from tests.support.mock import patch


@pytest.fixture
def configure_loader_modules():
    return {pecl: {}}


def test_install():
    """
    Test to installs one or several pecl extensions.
    """
    with patch.object(pecl, "_pecl", return_value="A"):
        assert pecl.install("fuse", force=True) == "A"

        assert not pecl.install("fuse")

        with patch.object(pecl, "list_", return_value={"A": ["A", "B"]}):
            assert pecl.install(["A", "B"])


def test_uninstall():
    """
    Test to uninstall one or several pecl extensions.
    """
    with patch.object(pecl, "_pecl", return_value="A"):
        assert pecl.uninstall("fuse") == "A"


def test_update():
    """
    Test to update one or several pecl extensions.
    """
    with patch.object(pecl, "_pecl", return_value="A"):
        assert pecl.update("fuse") == "A"


def test_list_():
    """
    Test to list installed pecl extensions.
    """
    with patch.object(pecl, "_pecl", return_value="A\nB"):
        assert pecl.list_("channel") == {}



"""
Library for interacting with Pushover API

.. versionadded:: 2016.3.0
"""

import http.client
import logging
from urllib.parse import urlencode, urljoin

import salt.utils.http
from salt.version import __version__

log = logging.getLogger(__name__)


def query(
    function,
    token=None,
    api_version="1",
    method="POST",
    header_dict=None,
    data=None,
    query_params=None,
    opts=None,
):
    """
    PushOver object method function to construct and execute on the API URL.

    :param token:       The PushOver api key.
    :param api_version: The PushOver API version to use, defaults to version 1.
    :param function:    The PushOver api function to perform.
    :param method:      The HTTP method, e.g. GET or POST.
    :param data:        The data to be sent for POST method.
    :return:            The json response from the API call or False.
    """

    ret = {"message": "", "res": True}

    pushover_functions = {
        "message": {"request": "messages.json", "response": "status"},
        "validate_user": {"request": "users/validate.json", "response": "status"},
        "validate_sound": {"request": "sounds.json", "response": "status"},
    }

    api_url = "https://api.pushover.net"
    base_url = urljoin(api_url, api_version + "/")
    path = pushover_functions.get(function).get("request")
    url = urljoin(base_url, path, False)

    if not query_params:
        query_params = {}

    decode = True
    if method == "DELETE":
        decode = False

    result = salt.utils.http.query(
        url,
        method,
        params=query_params,
        data=data,
        header_dict=header_dict,
        decode=decode,
        decode_type="json",
        text=True,
        status=True,
        cookies=True,
        persist_session=True,
        opts=opts,
    )

    if result.get("status", None) == http.client.OK:
        response = pushover_functions.get(function).get("response")
        if response in result and result[response] == 0:
            ret["res"] = False
        ret["message"] = result
        return ret
    else:
        try:
            if "response" in result and result[response] == 0:
                ret["res"] = False
            ret["message"] = result
        except ValueError:
            ret["res"] = False
            ret["message"] = result
        return ret


def validate_sound(sound, token):
    """
    Send a message to a Pushover user or group.
    :param sound:       The sound that we want to verify
    :param token:       The PushOver token.
    """
    ret = {"message": "Sound is invalid", "res": False}
    parameters = dict()
    parameters["token"] = token

    response = query(function="validate_sound", method="GET", query_params=parameters)

    if response["res"]:
        if "message" in response:
            _message = response.get("message", "")
            if "status" in _message:
                if _message.get("dict", {}).get("status", "") == 1:
                    sounds = _message.get("dict", {}).get("sounds", "")
                    if sound in sounds:
                        ret["message"] = "Valid sound {}.".format(sound)
                        ret["res"] = True
                    else:
                        ret["message"] = "Warning: {} not a valid sound.".format(sound)
                        ret["res"] = False
                else:
                    ret["message"] = "".join(_message.get("dict", {}).get("errors"))
    return ret


def validate_user(user, device, token):
    """
    Send a message to a Pushover user or group.
    :param user:        The user or group name, either will work.
    :param device:      The device for the user.
    :param token:       The PushOver token.
    """
    res = {"message": "User key is invalid", "result": False}

    parameters = dict()
    parameters["user"] = user
    parameters["token"] = token
    if device:
        parameters["device"] = device

    response = query(
        function="validate_user",
        method="POST",
        header_dict={"Content-Type": "application/x-www-form-urlencoded"},
        data=urlencode(parameters),
    )

    if response["res"]:
        if "message" in response:
            _message = response.get("message", "")
            if "status" in _message:
                if _message.get("dict", {}).get("status", None) == 1:
                    res["result"] = True
                    res["message"] = "User key is valid."
                else:
                    res["result"] = False
                    res["message"] = "".join(_message.get("dict", {}).get("errors"))
    return res



"""
This roster resolves hostname in a pdsh/clustershell style.

:depends: clustershell, https://github.com/cea-hpc/clustershell

When you want to use host globs for target matching, use ``--roster clustershell``. For example:

.. code-block:: bash

    salt-ssh --roster clustershell 'server_[1-10,21-30],test_server[5,7,9]' test.ping

"""


import copy
import socket

REQ_ERROR = None
try:
    from ClusterShell.NodeSet import NodeSet
except (ImportError, OSError) as e:
    REQ_ERROR = "ClusterShell import error, perhaps missing python ClusterShell package"


def __virtual__():
    return (REQ_ERROR is None, REQ_ERROR)


def targets(tgt, tgt_type="glob", **kwargs):
    """
    Return the targets
    """
    ret = {}
    ports = __opts__["ssh_scan_ports"]
    if not isinstance(ports, list):
        # Comma-separate list of integers
        ports = list(map(int, str(ports).split(",")))

    hosts = list(NodeSet(tgt))
    host_addrs = {h: socket.gethostbyname(h) for h in hosts}

    for host, addr in host_addrs.items():
        addr = str(addr)
        ret[host] = copy.deepcopy(__opts__.get("roster_defaults", {}))
        for port in ports:
            try:
                sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                sock.settimeout(float(__opts__["ssh_scan_timeout"]))
                sock.connect((addr, port))
                sock.shutdown(socket.SHUT_RDWR)
                sock.close()
                ret[host].update({"host": addr, "port": port})
            except OSError:
                pass
    return ret



"""
Manage SQS Queues

Create and destroy SQS queues. Be aware that this interacts with Amazon's
services, and so may incur charges.

This module uses the awscli tool provided by Amazon. This can be downloaded
from pip. Also check the documentation for awscli for configuration
information.

.. code-block:: yaml

    myqueue:
        aws_sqs.exists:
            - region: eu-west-1
"""


def __virtual__():
    """
    Only load if aws is available.
    """
    if __salt__["cmd.has_exec"]("aws"):
        return "aws_sqs"
    return (False, "aws command not found")


def exists(name, region, user=None, opts=False):
    """
    Ensure the SQS queue exists.

    name
        Name of the SQS queue.

    region
        Region to create the queue

    user
        Name of the user performing the SQS operations

    opts
        Include additional arguments and options to the aws command line
    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    does_exist = __salt__["aws_sqs.queue_exists"](name, region, opts, user)

    if not does_exist:
        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = "AWS SQS queue {} is set to be created".format(name)
            return ret
        created = __salt__["aws_sqs.create_queue"](name, region, opts, user)
        if created["retcode"] == 0:
            ret["changes"]["new"] = created["stdout"]
        else:
            ret["result"] = False
            ret["comment"] = created["stderr"]

    else:
        ret["comment"] = "{} exists in {}".format(name, region)

    return ret


def absent(name, region, user=None, opts=False):
    """
    Remove the named SQS queue if it exists.

    name
        Name of the SQS queue.

    region
        Region to remove the queue from

    user
        Name of the user performing the SQS operations

    opts
        Include additional arguments and options to the aws command line
    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    does_exist = __salt__["aws_sqs.queue_exists"](name, region, opts, user)

    if does_exist:
        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = "AWS SQS queue {} is set to be removed".format(name)
            return ret
        removed = __salt__["aws_sqs.delete_queue"](name, region, opts, user)
        if removed["retcode"] == 0:
            ret["changes"]["removed"] = removed["stdout"]
        else:
            ret["result"] = False
            ret["comment"] = removed["stderr"]
    else:
        ret["comment"] = "{} does not exist in {}".format(name, region)

    return ret



import logging

import pytest

import salt.sdb.etcd_db as etcd_db
from salt.utils.etcd_util import get_conn
from tests.support.pytest.etcd import *  # pylint: disable=wildcard-import,unused-wildcard-import

pytest.importorskip("docker")

log = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.skip_if_binaries_missing("docker", "dockerd", check_all=False),
]


@pytest.fixture(scope="module")
def etcd_client(minion_opts, profile_name):
    return get_conn(minion_opts, profile=profile_name)


@pytest.fixture(scope="module")
def prefix():
    return "/salt/sdb/test"


@pytest.fixture(autouse=True)
def cleanup_prefixed_entries(etcd_client, prefix):
    """
    Cleanup after each test to ensure a consistent starting state.
    """
    try:
        assert etcd_client.get(prefix, recurse=True) is None
        yield
    finally:
        etcd_client.delete(prefix, recurse=True)


def test_basic_operations(etcd_profile, prefix, profile_name):
    """
    Ensure we can do the basic CRUD operations available in sdb.etcd_db
    """
    assert (
        etcd_db.set_("{}/1".format(prefix), "one", profile=etcd_profile[profile_name])
        == "one"
    )
    etcd_db.delete("{}/1".format(prefix), profile=etcd_profile[profile_name])
    assert (
        etcd_db.get("{}/1".format(prefix), profile=etcd_profile[profile_name]) is None
    )



"""
    :codeauthor: Rahul Handay <rahulha@saltstack.com>
"""


import pytest

import salt.modules.saltcloudmod as saltcloudmod
import salt.utils.json
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {saltcloudmod: {}}


def test_create():
    """
    Test if create the named vm
    """
    mock = MagicMock(return_value="""{"foo": "bar"}""")
    mock_json_loads = MagicMock(side_effect=ValueError())
    with patch.dict(saltcloudmod.__salt__, {"cmd.run_stdout": mock}):
        assert saltcloudmod.create("webserver", "rackspace_centos_512")

        with patch.object(salt.utils.json, "loads", mock_json_loads):
            assert saltcloudmod.create("webserver", "rackspace_centos_512") == {}



# pylint: skip-file
from __future__ import absolute_import, division, print_function

import functools
import sys
import traceback

from salt.ext.tornado.concurrent import Future
from salt.ext.tornado import gen
from salt.ext.tornado.httpclient import HTTPError, HTTPRequest
from salt.ext.tornado.log import gen_log, app_log
from salt.ext.tornado.template import DictLoader
from salt.ext.tornado.testing import AsyncHTTPTestCase, gen_test, bind_unused_port, ExpectLog
from salt.ext.tornado.test.util import unittest, skipBefore35, exec_test
from salt.ext.tornado.web import Application, RequestHandler

try:
    import salt.ext.tornado.websocket  # noqa
    from salt.ext.tornado.util import _websocket_mask_python
except ImportError:
    # The unittest module presents misleading errors on ImportError
    # (it acts as if websocket_test could not be found, hiding the underlying
    # error).  If we get an ImportError here (which could happen due to
    # TORNADO_EXTENSION=1), print some extra information before failing.
    traceback.print_exc()
    raise

from salt.ext.tornado.websocket import WebSocketHandler, websocket_connect, WebSocketError

try:
    from salt.ext.tornado import speedups
except ImportError:
    speedups = None


class TestWebSocketHandler(WebSocketHandler):
    """Base class for testing handlers that exposes the on_close event.

    This allows for deterministic cleanup of the associated socket.
    """
    def initialize(self, close_future, compression_options=None):
        self.close_future = close_future
        self.compression_options = compression_options

    def get_compression_options(self):
        return self.compression_options

    def on_close(self):
        self.close_future.set_result((self.close_code, self.close_reason))


class EchoHandler(TestWebSocketHandler):
    def on_message(self, message):
        self.write_message(message, isinstance(message, bytes))


class ErrorInOnMessageHandler(TestWebSocketHandler):
    def on_message(self, message):
        1 / 0


class HeaderHandler(TestWebSocketHandler):
    def open(self):
        methods_to_test = [
            functools.partial(self.write, 'This should not work'),
            functools.partial(self.redirect, 'http://localhost/elsewhere'),
            functools.partial(self.set_header, 'X-Test', ''),
            functools.partial(self.set_cookie, 'Chocolate', 'Chip'),
            functools.partial(self.set_status, 503),
            self.flush,
            self.finish,
        ]
        for method in methods_to_test:
            try:
                # In a websocket context, many RequestHandler methods
                # raise RuntimeErrors.
                method()
                raise Exception("did not get expected exception")
            except RuntimeError:
                pass
        self.write_message(self.request.headers.get('X-Test', ''))


class HeaderEchoHandler(TestWebSocketHandler):
    def set_default_headers(self):
        self.set_header("X-Extra-Response-Header", "Extra-Response-Value")

    def prepare(self):
        for k, v in self.request.headers.get_all():
            if k.lower().startswith('x-test'):
                self.set_header(k, v)


class NonWebSocketHandler(RequestHandler):
    def get(self):
        self.write('ok')


class CloseReasonHandler(TestWebSocketHandler):
    def open(self):
        self.on_close_called = False
        self.close(1001, "goodbye")


class AsyncPrepareHandler(TestWebSocketHandler):
    @gen.coroutine
    def prepare(self):
        yield gen.moment

    def on_message(self, message):
        self.write_message(message)


class PathArgsHandler(TestWebSocketHandler):
    def open(self, arg):
        self.write_message(arg)


class CoroutineOnMessageHandler(TestWebSocketHandler):
    def initialize(self, close_future, compression_options=None):
        super(CoroutineOnMessageHandler, self).initialize(close_future,
                                                          compression_options)
        self.sleeping = 0

    @gen.coroutine
    def on_message(self, message):
        if self.sleeping > 0:
            self.write_message('another coroutine is already sleeping')
        self.sleeping += 1
        yield gen.sleep(0.01)
        self.sleeping -= 1
        self.write_message(message)


class RenderMessageHandler(TestWebSocketHandler):
    def on_message(self, message):
        self.write_message(self.render_string('message.html', message=message))


class WebSocketBaseTestCase(AsyncHTTPTestCase):
    @gen.coroutine
    def ws_connect(self, path, **kwargs):
        ws = yield websocket_connect(
            'ws://127.0.0.1:%d%s' % (self.get_http_port(), path),
            **kwargs)
        raise gen.Return(ws)

    @gen.coroutine
    def close(self, ws):
        """Close a websocket connection and wait for the server side.

        If we don't wait here, there are sometimes leak warnings in the
        tests.
        """
        ws.close()
        yield self.close_future


class WebSocketTest(WebSocketBaseTestCase):
    def get_app(self):
        self.close_future = Future()
        return Application([
            ('/echo', EchoHandler, dict(close_future=self.close_future)),
            ('/non_ws', NonWebSocketHandler),
            ('/header', HeaderHandler, dict(close_future=self.close_future)),
            ('/header_echo', HeaderEchoHandler,
             dict(close_future=self.close_future)),
            ('/close_reason', CloseReasonHandler,
             dict(close_future=self.close_future)),
            ('/error_in_on_message', ErrorInOnMessageHandler,
             dict(close_future=self.close_future)),
            ('/async_prepare', AsyncPrepareHandler,
             dict(close_future=self.close_future)),
            ('/path_args/(.*)', PathArgsHandler,
             dict(close_future=self.close_future)),
            ('/coroutine', CoroutineOnMessageHandler,
             dict(close_future=self.close_future)),
            ('/render', RenderMessageHandler,
             dict(close_future=self.close_future)),
        ], template_loader=DictLoader({
            'message.html': '<b>{{ message }}</b>',
        }))

    def tearDown(self):
        super(WebSocketTest, self).tearDown()
        RequestHandler._template_loaders.clear()

    def test_http_request(self):
        # WS server, HTTP client.
        response = self.fetch('/echo')
        self.assertEqual(response.code, 400)

    def test_missing_websocket_key(self):
        response = self.fetch('/echo',
                              headers={'Connection': 'Upgrade',
                                       'Upgrade': 'WebSocket',
                                       'Sec-WebSocket-Version': '13'})
        self.assertEqual(response.code, 400)

    def test_bad_websocket_version(self):
        response = self.fetch('/echo',
                              headers={'Connection': 'Upgrade',
                                       'Upgrade': 'WebSocket',
                                       'Sec-WebSocket-Version': '12'})
        self.assertEqual(response.code, 426)

    @gen_test
    def test_websocket_gen(self):
        ws = yield self.ws_connect('/echo')
        yield ws.write_message('hello')
        response = yield ws.read_message()
        self.assertEqual(response, 'hello')
        yield self.close(ws)

    def test_websocket_callbacks(self):
        websocket_connect(
            'ws://127.0.0.1:%d/echo' % self.get_http_port(),
            io_loop=self.io_loop, callback=self.stop)
        ws = self.wait().result()
        ws.write_message('hello')
        ws.read_message(self.stop)
        response = self.wait().result()
        self.assertEqual(response, 'hello')
        self.close_future.add_done_callback(lambda f: self.stop())
        ws.close()
        self.wait()

    @gen_test
    def test_binary_message(self):
        ws = yield self.ws_connect('/echo')
        ws.write_message(b'hello \xe9', binary=True)
        response = yield ws.read_message()
        self.assertEqual(response, b'hello \xe9')
        yield self.close(ws)

    @gen_test
    def test_unicode_message(self):
        ws = yield self.ws_connect('/echo')
        ws.write_message(u'hello \u00e9')
        response = yield ws.read_message()
        self.assertEqual(response, u'hello \u00e9')
        yield self.close(ws)

    @gen_test
    def test_render_message(self):
        ws = yield self.ws_connect('/render')
        ws.write_message('hello')
        response = yield ws.read_message()
        self.assertEqual(response, '<b>hello</b>')
        yield self.close(ws)

    @gen_test
    def test_error_in_on_message(self):
        ws = yield self.ws_connect('/error_in_on_message')
        ws.write_message('hello')
        with ExpectLog(app_log, "Uncaught exception"):
            response = yield ws.read_message()
        self.assertIs(response, None)
        yield self.close(ws)

    @gen_test
    def test_websocket_http_fail(self):
        with self.assertRaises(HTTPError) as cm:
            yield self.ws_connect('/notfound')
        self.assertEqual(cm.exception.code, 404)

    @gen_test
    def test_websocket_http_success(self):
        with self.assertRaises(WebSocketError):
            yield self.ws_connect('/non_ws')

    @gen_test
    def test_websocket_network_fail(self):
        sock, port = bind_unused_port()
        sock.close()
        with self.assertRaises(IOError):
            with ExpectLog(gen_log, ".*"):
                yield websocket_connect(
                    'ws://127.0.0.1:%d/' % port,
                    io_loop=self.io_loop,
                    connect_timeout=3600)

    @gen_test
    def test_websocket_close_buffered_data(self):
        ws = yield websocket_connect(
            'ws://127.0.0.1:%d/echo' % self.get_http_port())
        ws.write_message('hello')
        ws.write_message('world')
        # Close the underlying stream.
        ws.stream.close()
        yield self.close_future

    @gen_test
    def test_websocket_headers(self):
        # Ensure that arbitrary headers can be passed through websocket_connect.
        ws = yield websocket_connect(
            HTTPRequest('ws://127.0.0.1:%d/header' % self.get_http_port(),
                        headers={'X-Test': 'hello'}))
        response = yield ws.read_message()
        self.assertEqual(response, 'hello')
        yield self.close(ws)

    @gen_test
    def test_websocket_header_echo(self):
        # Ensure that headers can be returned in the response.
        # Specifically, that arbitrary headers passed through websocket_connect
        # can be returned.
        ws = yield websocket_connect(
            HTTPRequest('ws://127.0.0.1:%d/header_echo' % self.get_http_port(),
                        headers={'X-Test-Hello': 'hello'}))
        self.assertEqual(ws.headers.get('X-Test-Hello'), 'hello')
        self.assertEqual(ws.headers.get('X-Extra-Response-Header'), 'Extra-Response-Value')
        yield self.close(ws)

    @gen_test
    def test_server_close_reason(self):
        ws = yield self.ws_connect('/close_reason')
        msg = yield ws.read_message()
        # A message of None means the other side closed the connection.
        self.assertIs(msg, None)
        self.assertEqual(ws.close_code, 1001)
        self.assertEqual(ws.close_reason, "goodbye")
        # The on_close callback is called no matter which side closed.
        code, reason = yield self.close_future
        # The client echoed the close code it received to the server,
        # so the server's close code (returned via close_future) is
        # the same.
        self.assertEqual(code, 1001)

    @gen_test
    def test_client_close_reason(self):
        ws = yield self.ws_connect('/echo')
        ws.close(1001, 'goodbye')
        code, reason = yield self.close_future
        self.assertEqual(code, 1001)
        self.assertEqual(reason, 'goodbye')

    @gen_test
    def test_async_prepare(self):
        # Previously, an async prepare method triggered a bug that would
        # result in a timeout on test shutdown (and a memory leak).
        ws = yield self.ws_connect('/async_prepare')
        ws.write_message('hello')
        res = yield ws.read_message()
        self.assertEqual(res, 'hello')

    @gen_test
    def test_path_args(self):
        ws = yield self.ws_connect('/path_args/hello')
        res = yield ws.read_message()
        self.assertEqual(res, 'hello')

    @gen_test
    def test_coroutine(self):
        ws = yield self.ws_connect('/coroutine')
        # Send both messages immediately, coroutine must process one at a time.
        yield ws.write_message('hello1')
        yield ws.write_message('hello2')
        res = yield ws.read_message()
        self.assertEqual(res, 'hello1')
        res = yield ws.read_message()
        self.assertEqual(res, 'hello2')

    @gen_test
    def test_check_origin_valid_no_path(self):
        port = self.get_http_port()

        url = 'ws://127.0.0.1:%d/echo' % port
        headers = {'Origin': 'http://127.0.0.1:%d' % port}

        ws = yield websocket_connect(HTTPRequest(url, headers=headers),
                                     io_loop=self.io_loop)
        ws.write_message('hello')
        response = yield ws.read_message()
        self.assertEqual(response, 'hello')
        yield self.close(ws)

    @gen_test
    def test_check_origin_valid_with_path(self):
        port = self.get_http_port()

        url = 'ws://127.0.0.1:%d/echo' % port
        headers = {'Origin': 'http://127.0.0.1:%d/something' % port}

        ws = yield websocket_connect(HTTPRequest(url, headers=headers),
                                     io_loop=self.io_loop)
        ws.write_message('hello')
        response = yield ws.read_message()
        self.assertEqual(response, 'hello')
        yield self.close(ws)

    @gen_test
    def test_check_origin_invalid_partial_url(self):
        port = self.get_http_port()

        url = 'ws://127.0.0.1:%d/echo' % port
        headers = {'Origin': '127.0.0.1:%d' % port}

        with self.assertRaises(HTTPError) as cm:
            yield websocket_connect(HTTPRequest(url, headers=headers),
                                    io_loop=self.io_loop)
        self.assertEqual(cm.exception.code, 403)

    @gen_test
    def test_check_origin_invalid(self):
        port = self.get_http_port()

        url = 'ws://127.0.0.1:%d/echo' % port
        # Host is 127.0.0.1, which should not be accessible from some other
        # domain
        headers = {'Origin': 'http://somewhereelse.com'}

        with self.assertRaises(HTTPError) as cm:
            yield websocket_connect(HTTPRequest(url, headers=headers),
                                    io_loop=self.io_loop)

        self.assertEqual(cm.exception.code, 403)

    @gen_test
    def test_check_origin_invalid_subdomains(self):
        port = self.get_http_port()

        url = 'ws://localhost:%d/echo' % port
        # Subdomains should be disallowed by default.  If we could pass a
        # resolver to websocket_connect we could test sibling domains as well.
        headers = {'Origin': 'http://subtenant.localhost'}

        with self.assertRaises(HTTPError) as cm:
            yield websocket_connect(HTTPRequest(url, headers=headers),
                                    io_loop=self.io_loop)

        self.assertEqual(cm.exception.code, 403)


if sys.version_info >= (3, 5):
    NativeCoroutineOnMessageHandler = exec_test(globals(), locals(), """
class NativeCoroutineOnMessageHandler(TestWebSocketHandler):
    def initialize(self, close_future, compression_options=None):
        super().initialize(close_future, compression_options)
        self.sleeping = 0

    async def on_message(self, message):
        if self.sleeping > 0:
            self.write_message('another coroutine is already sleeping')
        self.sleeping += 1
        await gen.sleep(0.01)
        self.sleeping -= 1
        self.write_message(message)""")['NativeCoroutineOnMessageHandler']


class WebSocketNativeCoroutineTest(WebSocketBaseTestCase):
    def get_app(self):
        self.close_future = Future()
        return Application([
            ('/native', NativeCoroutineOnMessageHandler,
             dict(close_future=self.close_future))])

    @skipBefore35
    @gen_test
    def test_native_coroutine(self):
        ws = yield self.ws_connect('/native')
        # Send both messages immediately, coroutine must process one at a time.
        yield ws.write_message('hello1')
        yield ws.write_message('hello2')
        res = yield ws.read_message()
        self.assertEqual(res, 'hello1')
        res = yield ws.read_message()
        self.assertEqual(res, 'hello2')


class CompressionTestMixin(object):
    MESSAGE = 'Hello world. Testing 123 123'

    def get_app(self):
        self.close_future = Future()
        return Application([
            ('/echo', EchoHandler, dict(
                close_future=self.close_future,
                compression_options=self.get_server_compression_options())),
        ])

    def get_server_compression_options(self):
        return None

    def get_client_compression_options(self):
        return None

    @gen_test
    def test_message_sizes(self):
        ws = yield self.ws_connect(
            '/echo',
            compression_options=self.get_client_compression_options())
        # Send the same message three times so we can measure the
        # effect of the context_takeover options.
        for i in range(3):
            ws.write_message(self.MESSAGE)
            response = yield ws.read_message()
            self.assertEqual(response, self.MESSAGE)
        self.assertEqual(ws.protocol._message_bytes_out, len(self.MESSAGE) * 3)
        self.assertEqual(ws.protocol._message_bytes_in, len(self.MESSAGE) * 3)
        self.verify_wire_bytes(ws.protocol._wire_bytes_in,
                               ws.protocol._wire_bytes_out)
        yield self.close(ws)


class UncompressedTestMixin(CompressionTestMixin):
    """Specialization of CompressionTestMixin when we expect no compression."""
    def verify_wire_bytes(self, bytes_in, bytes_out):
        # Bytes out includes the 4-byte mask key per message.
        self.assertEqual(bytes_out, 3 * (len(self.MESSAGE) + 6))
        self.assertEqual(bytes_in, 3 * (len(self.MESSAGE) + 2))


class NoCompressionTest(UncompressedTestMixin, WebSocketBaseTestCase):
    pass


# If only one side tries to compress, the extension is not negotiated.
class ServerOnlyCompressionTest(UncompressedTestMixin, WebSocketBaseTestCase):
    def get_server_compression_options(self):
        return {}


class ClientOnlyCompressionTest(UncompressedTestMixin, WebSocketBaseTestCase):
    def get_client_compression_options(self):
        return {}


class DefaultCompressionTest(CompressionTestMixin, WebSocketBaseTestCase):
    def get_server_compression_options(self):
        return {}

    def get_client_compression_options(self):
        return {}

    def verify_wire_bytes(self, bytes_in, bytes_out):
        self.assertLess(bytes_out, 3 * (len(self.MESSAGE) + 6))
        self.assertLess(bytes_in, 3 * (len(self.MESSAGE) + 2))
        # Bytes out includes the 4 bytes mask key per message.
        self.assertEqual(bytes_out, bytes_in + 12)


class MaskFunctionMixin(object):
    # Subclasses should define self.mask(mask, data)
    def test_mask(self):
        self.assertEqual(self.mask(b'abcd', b''), b'')
        self.assertEqual(self.mask(b'abcd', b'b'), b'\x03')
        self.assertEqual(self.mask(b'abcd', b'54321'), b'TVPVP')
        self.assertEqual(self.mask(b'ZXCV', b'98765432'), b'c`t`olpd')
        # Include test cases with \x00 bytes (to ensure that the C
        # extension isn't depending on null-terminated strings) and
        # bytes with the high bit set (to smoke out signedness issues).
        self.assertEqual(self.mask(b'\x00\x01\x02\x03',
                                   b'\xff\xfb\xfd\xfc\xfe\xfa'),
                         b'\xff\xfa\xff\xff\xfe\xfb')
        self.assertEqual(self.mask(b'\xff\xfb\xfd\xfc',
                                   b'\x00\x01\x02\x03\x04\x05'),
                         b'\xff\xfa\xff\xff\xfb\xfe')


class PythonMaskFunctionTest(MaskFunctionMixin, unittest.TestCase):
    def mask(self, mask, data):
        return _websocket_mask_python(mask, data)


@unittest.skipIf(speedups is None, "tornado.speedups module not present")
class CythonMaskFunctionTest(MaskFunctionMixin, unittest.TestCase):
    def mask(self, mask, data):
        return speedups.websocket_mask(mask, data)


class ServerPeriodicPingTest(WebSocketBaseTestCase):
    def get_app(self):
        class PingHandler(TestWebSocketHandler):
            def on_pong(self, data):
                self.write_message("got pong")

        self.close_future = Future()
        return Application([
            ('/', PingHandler, dict(close_future=self.close_future)),
        ], websocket_ping_interval=0.01)

    @gen_test
    def test_server_ping(self):
        ws = yield self.ws_connect('/')
        for i in range(3):
            response = yield ws.read_message()
            self.assertEqual(response, "got pong")
        yield self.close(ws)
        # TODO: test that the connection gets closed if ping responses stop.


class ClientPeriodicPingTest(WebSocketBaseTestCase):
    def get_app(self):
        class PingHandler(TestWebSocketHandler):
            def on_ping(self, data):
                self.write_message("got ping")

        self.close_future = Future()
        return Application([
            ('/', PingHandler, dict(close_future=self.close_future)),
        ])

    @gen_test
    def test_client_ping(self):
        ws = yield self.ws_connect('/', ping_interval=0.01)
        for i in range(3):
            response = yield ws.read_message()
            self.assertEqual(response, "got ping")
        yield self.close(ws)
        # TODO: test that the connection gets closed if ping responses stop.


class MaxMessageSizeTest(WebSocketBaseTestCase):
    def get_app(self):
        self.close_future = Future()
        return Application([
            ('/', EchoHandler, dict(close_future=self.close_future)),
        ], websocket_max_message_size=1024)

    @gen_test
    def test_large_message(self):
        ws = yield self.ws_connect('/')

        # Write a message that is allowed.
        msg = 'a' * 1024
        ws.write_message(msg)
        resp = yield ws.read_message()
        self.assertEqual(resp, msg)

        # Write a message that is too large.
        ws.write_message(msg + 'b')
        resp = yield ws.read_message()
        # A message of None means the other side closed the connection.
        self.assertIs(resp, None)
        self.assertEqual(ws.close_code, 1009)
        self.assertEqual(ws.close_reason, "message too big")
        # TODO: Needs tests of messages split over multiple
        # continuation frames.



import pathlib

import attr
import pytest

import salt.channel.server
import salt.ext.tornado.gen
import salt.transport.ipc
import salt.utils.platform
from salt.ext.tornado import locks

pytestmark = [
    # Windows does not support POSIX IPC
    pytest.mark.skip_on_windows,
]


@attr.s(frozen=True, slots=True)
class PayloadHandler:
    payloads = attr.ib(init=False, default=attr.Factory(list))

    async def handle_payload(self, payload, reply_func):
        self.payloads.append(payload)
        await reply_func(payload)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.payloads.clear()


@attr.s(frozen=True, slots=True)
class IPCTester:
    io_loop = attr.ib()
    socket_path = attr.ib()
    publisher = attr.ib()
    subscriber = attr.ib()
    payloads = attr.ib(default=attr.Factory(list))
    payload_ack = attr.ib(default=attr.Factory(locks.Condition))

    @subscriber.default
    def _subscriber_default(self):
        return salt.transport.ipc.IPCMessageSubscriber(
            self.socket_path,
            io_loop=self.io_loop,
        )

    @publisher.default
    def _publisher_default(self):
        return salt.transport.ipc.IPCMessagePublisher(
            {"ipc_write_buffer": 0},
            self.socket_path,
            io_loop=self.io_loop,
        )

    async def handle_payload(self, payload, reply_func):
        self.payloads.append(payload)
        await reply_func(payload)
        self.payload_ack.notify()

    def new_client(self):
        return IPCTester(
            io_loop=self.io_loop,
            socket_path=self.socket_path,
            server=self.server,
            payloads=self.payloads,
            payload_ack=self.payload_ack,
        )

    async def publish(self, payload, timeout=60):
        self.publisher.publish(payload)

    async def read(self, timeout=60):
        ret = await self.subscriber.read(timeout)
        return ret

    def __enter__(self):
        self.publisher.start()
        self.io_loop.add_callback(self.subscriber.connect)
        return self

    def __exit__(self, *args):
        self.subscriber.close()
        self.publisher.close()


@pytest.fixture
def ipc_socket_path(tmp_path):
    if salt.utils.platform.is_darwin():
        # A shorter path so that we don't hit the AF_UNIX path too long
        tmp_path = pathlib.Path("/tmp").resolve()
    _socket_path = tmp_path / "ipc-test.ipc"
    try:
        yield _socket_path
    finally:
        if _socket_path.exists():
            _socket_path.unlink()


@pytest.fixture
def channel(io_loop, ipc_socket_path):
    _ipc_tester = IPCTester(io_loop=io_loop, socket_path=str(ipc_socket_path))
    with _ipc_tester:
        yield _ipc_tester


async def test_basic_send(channel):
    msg = {"foo": "bar", "stop": True}
    # XXX: IPCClient connect and connected methods need to be cleaned up as
    # this should not be needed.
    while not channel.subscriber._connecting_future.done():
        await salt.ext.tornado.gen.sleep(0.01)
    while not channel.subscriber.connected():
        await salt.ext.tornado.gen.sleep(0.01)
    assert channel.subscriber.connected()
    await channel.publish(msg)
    ret = await channel.read()
    assert ret == msg



"""
Return data to a mongodb server

Required python modules: pymongo


This returner will send data from the minions to a MongoDB server. To
configure the settings for your MongoDB server, add the following lines
to the minion config files.

.. code-block:: yaml

    mongo.db: <database name>
    mongo.host: <server ip address>
    mongo.user: <MongoDB username>
    mongo.password: <MongoDB user password>
    mongo.port: 27017

Alternative configuration values can be used by prefacing the configuration.
Any values not found in the alternative configuration will be pulled from
the default location.

.. code-block:: yaml

    alternative.mongo.db: <database name>
    alternative.mongo.host: <server ip address>
    alternative.mongo.user: <MongoDB username>
    alternative.mongo.password: <MongoDB user password>
    alternative.mongo.port: 27017

To use the mongo returner, append '--return mongo' to the salt command.

.. code-block:: bash

    salt '*' test.ping --return mongo_return

To use the alternative configuration, append '--return_config alternative' to the salt command.

.. versionadded:: 2015.5.0

.. code-block:: bash

    salt '*' test.ping --return mongo_return --return_config alternative

To override individual configuration items, append --return_kwargs '{"key:": "value"}' to the salt command.

.. versionadded:: 2016.3.0

.. code-block:: bash

    salt '*' test.ping --return mongo --return_kwargs '{"db": "another-salt"}'

To override individual configuration items, append --return_kwargs '{"key:": "value"}' to the salt command.

.. versionadded:: 2016.3.0

.. code-block:: bash

    salt '*' test.ping --return mongo --return_kwargs '{"db": "another-salt"}'

"""

import logging

import salt.returners
import salt.utils.jid
from salt.utils.versions import Version

try:
    import pymongo

    PYMONGO_VERSION = Version(pymongo.version)
    HAS_PYMONGO = True
except ImportError:
    HAS_PYMONGO = False


log = logging.getLogger(__name__)

# Define the module's virtual name
# currently only used iby _get_options
__virtualname__ = "mongo"


def __virtual__():
    if not HAS_PYMONGO:
        return False, "Could not import mongo returner; pymongo is not installed."
    return "mongo_return"


def _remove_dots(src):
    """
    Remove dots from the given data structure
    """
    output = {}
    for key, val in src.items():
        if isinstance(val, dict):
            val = _remove_dots(val)
        output[key.replace(".", "-")] = val
    return output


def _get_options(ret):
    """
    Get the monogo_return options from salt.
    """
    attrs = {
        "host": "host",
        "port": "port",
        "db": "db",
        "user": "user",
        "password": "password",
        "indexes": "indexes",
    }

    _options = salt.returners.get_returner_options(
        __virtualname__, ret, attrs, __salt__=__salt__, __opts__=__opts__
    )
    return _options


def _get_conn(ret):
    """
    Return a mongodb connection object
    """
    _options = _get_options(ret)

    host = _options.get("host")
    port = _options.get("port")
    db_ = _options.get("db")
    user = _options.get("user")
    password = _options.get("password")
    indexes = _options.get("indexes", False)

    # at some point we should remove support for
    # pymongo versions < 2.3 until then there are
    # a bunch of these sections that need to be supported

    if PYMONGO_VERSION > Version("2.3"):
        conn = pymongo.MongoClient(host, port)
    else:
        conn = pymongo.Connection(host, port)
    mdb = conn[db_]

    if user and password:
        mdb.authenticate(user, password)

    if indexes:
        if PYMONGO_VERSION > Version("2.3"):
            mdb.saltReturns.create_index("minion")
            mdb.saltReturns.create_index("jid")

            mdb.jobs.create_index("jid")
        else:
            mdb.saltReturns.ensure_index("minion")
            mdb.saltReturns.ensure_index("jid")

            mdb.jobs.ensure_index("jid")

    return conn, mdb


def returner(ret):
    """
    Return data to a mongodb server
    """
    conn, mdb = _get_conn(ret)
    col = mdb[ret["id"]]

    if isinstance(ret["return"], dict):
        back = _remove_dots(ret["return"])
    else:
        back = ret["return"]

    if isinstance(ret, dict):
        full_ret = _remove_dots(ret)
    else:
        full_ret = ret

    log.debug(back)
    sdata = {
        "minion": ret["id"],
        "jid": ret["jid"],
        "return": back,
        "fun": ret["fun"],
        "full_ret": full_ret,
    }
    if "out" in ret:
        sdata["out"] = ret["out"]

    # save returns in the saltReturns collection in the json format:
    # { 'minion': <minion_name>, 'jid': <job_id>, 'return': <return info with dots removed>,
    #  'fun': <function>, 'full_ret': <unformatted return with dots removed>}

    # again we run into the issue with deprecated code from previous versions

    if PYMONGO_VERSION > Version("2.3"):
        # using .copy() to ensure original data for load is unchanged
        mdb.saltReturns.insert_one(sdata.copy())
    else:
        mdb.saltReturns.insert(sdata.copy())


def get_jid(jid):
    """
    Return the return information associated with a jid
    """
    conn, mdb = _get_conn(ret=None)
    ret = {}
    rdata = mdb.saltReturns.find({"jid": jid}, {"_id": 0})
    if rdata:
        for data in rdata:
            minion = data["minion"]
            # return data in the format {<minion>: { <unformatted full return data>}}
            ret[minion] = data["full_ret"]
    return ret


def get_fun(fun):
    """
    Return the most recent jobs that have executed the named function
    """
    conn, mdb = _get_conn(ret=None)
    ret = {}
    rdata = mdb.saltReturns.find_one({"fun": fun}, {"_id": 0})
    if rdata:
        ret = rdata
    return ret


def prep_jid(nocache=False, passed_jid=None):  # pylint: disable=unused-argument
    """
    Do any work necessary to prepare a JID, including sending a custom id
    """
    return passed_jid if passed_jid is not None else salt.utils.jid.gen_jid(__opts__)


def save_minions(jid, minions, syndic_id=None):  # pylint: disable=unused-argument
    """
    Included for API consistency
    """



import pytest

import salt.modules.cryptdev as cryptdev
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules(minion_opts):
    return {cryptdev: {"__opts__": minion_opts}}


def test_active(caplog):
    with patch.dict(
        cryptdev.__salt__,
        {"cmd.run_stdout": MagicMock(return_value="my-device       (253, 1)\n")},
    ):
        assert cryptdev.active() == {
            "my-device": {
                "devname": "my-device",
                "major": "253",
                "minor": "1",
            }
        }

    # debien output when no devices setup.
    with patch.dict(cryptdev.__salt__, {"cmd.run_stdout": MagicMock(return_value="")}):
        caplog.clear()
        assert cryptdev.active() == {}
        assert "dmsetup output does not match expected format" in caplog.text

    # centos output of dmsetup when no devices setup.
    with patch.dict(
        cryptdev.__salt__,
        {"cmd.run_stdout": MagicMock(return_value="No devices found")},
    ):
        caplog.clear()
        assert cryptdev.active() == {}
        assert "dmsetup output does not match expected format" in caplog.text



import textwrap

import salt.modules.xfs as xfs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


@patch("salt.modules.xfs._get_mounts", MagicMock(return_value={}))
class XFSTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.xfs
    """

    def setup_loader_modules(self):
        return {xfs: {}}

    def test__blkid_output(self):
        """
        Test xfs._blkid_output when there is data
        """
        blkid_export = textwrap.dedent(
            """
            DEVNAME=/dev/sda1
            UUID=XXXXXXXX-XXXX-XXXX-XXXX-XXXXXXXXXXXX
            TYPE=xfs
            PARTUUID=YYYYYYYY-YY

            DEVNAME=/dev/sdb1
            PARTUUID=ZZZZZZZZ-ZZZZ-ZZZZ-ZZZZ-ZZZZZZZZZZZZ
            """
        )
        # We expect to find only data from /dev/sda1, nothig from
        # /dev/sdb1
        self.assertEqual(
            xfs._blkid_output(blkid_export),
            {
                "/dev/sda1": {
                    "label": None,
                    "partuuid": "YYYYYYYY-YY",
                    "uuid": "XXXXXXXX-XXXX-XXXX-XXXX-XXXXXXXXXXXX",
                }
            },
        )

    def test__parse_xfs_info(self):
        """
        Test parsing output from mkfs.xfs.
        """
        data = textwrap.dedent(
            """
            meta-data=/dev/vg00/testvol      isize=512    agcount=4, agsize=1310720 blks
                     =                       sectsz=4096  attr=2, projid32bit=1
                     =                       crc=1        finobt=1, sparse=1, rmapbt=0
                     =                       reflink=1
            data     =                       bsize=4096   blocks=5242880, imaxpct=25
                     =                       sunit=0      swidth=0 blks
            naming   =version 2              bsize=4096   ascii-ci=0, ftype=1
            log      =internal log           bsize=4096   blocks=2560, version=2
                     =                       sectsz=4096  sunit=1 blks, lazy-count=1
            realtime =none                   extsz=4096   blocks=0, rtextents=0
            Discarding blocks...Done.
            """
        )

        self.assertEqual(
            xfs._parse_xfs_info(data),
            {
                "meta-data": {
                    "section": "/dev/vg00/testvol",
                    "isize": "512",
                    "agcount": "4",
                    "agsize": "1310720 blks",
                    "sectsz": "4096",
                    "attr": "2",
                    "projid32bit": "1",
                    "crc": "1",
                    "finobt": "1",
                    "sparse": "1",
                    "rmapbt": "0",
                    "reflink": "1",
                },
                "data": {
                    "section": "data",
                    "bsize": "4096",
                    "blocks": "5242880",
                    "imaxpct": "25",
                    "sunit": "0",
                    "swidth": "0 blks",
                },
                "naming": {
                    "section": "version 2",
                    "bsize": "4096",
                    "ascii-ci": "0",
                    "ftype": "1",
                },
                "log": {
                    "section": "internal log",
                    "bsize": "4096",
                    "blocks": "2560",
                    "version": "2",
                    "sectsz": "4096",
                    "sunit": "1 blks",
                    "lazy-count": "1",
                },
                "realtime": {
                    "section": "none",
                    "extsz": "4096",
                    "blocks": "0",
                    "rtextents": "0",
                },
            },
        )



"""
    :codeauthor: Pedro Algarvio (pedro@algarvio.me)
    :copyright: Copyright 2017 by the SaltStack Team, see AUTHORS for more details.
    :license: Apache 2.0, see LICENSE for more details.


    tests.support.paths
    ~~~~~~~~~~~~~~~~~~~

    Tests related paths
"""

import logging
import os
import re
import sys
import tempfile

log = logging.getLogger(__name__)

SALT_CODE_DIR = os.path.join(
    os.path.dirname(
        os.path.dirname(
            os.path.dirname(
                os.path.dirname(os.path.normpath(os.path.abspath(__file__)))
            )
        )
    ),
    "salt",
)
TESTS_DIR = os.path.join(os.path.dirname(SALT_CODE_DIR), "tests")
if TESTS_DIR.startswith("//"):
    # Have we been given an initial double forward slash? Ditch it!
    TESTS_DIR = TESTS_DIR[1:]
if sys.platform.startswith("win"):
    TESTS_DIR = os.path.normcase(TESTS_DIR)
CODE_DIR = os.path.dirname(TESTS_DIR)
if sys.platform.startswith("win"):
    CODE_DIR = CODE_DIR.replace("\\", "\\\\")
UNIT_TEST_DIR = os.path.join(TESTS_DIR, "unit")
INTEGRATION_TEST_DIR = os.path.join(TESTS_DIR, "integration")

# Let's inject CODE_DIR so salt is importable if not there already
if TESTS_DIR in sys.path:
    sys.path.remove(TESTS_DIR)
if CODE_DIR in sys.path and sys.path[0] != CODE_DIR:
    sys.path.remove(CODE_DIR)
if CODE_DIR not in sys.path:
    sys.path.insert(0, CODE_DIR)
if TESTS_DIR not in sys.path:
    sys.path.insert(1, TESTS_DIR)

SYS_TMP_DIR = os.path.abspath(
    os.path.realpath(
        # Avoid ${TMPDIR} and gettempdir() on MacOS as they yield a base path too long
        # for unix sockets: ``error: AF_UNIX path too long``
        # Gentoo Portage prefers ebuild tests are rooted in ${TMPDIR}
        os.environ.get("TMPDIR", tempfile.gettempdir())
        if not sys.platform.startswith("darwin")
        else "/tmp"
    )
)
TMP = os.path.join(SYS_TMP_DIR, "salt-tests-tmpdir")
TMP_ROOT_DIR = os.path.join(TMP, "rootdir")
FILES = os.path.join(INTEGRATION_TEST_DIR, "files")
BASE_FILES = os.path.join(INTEGRATION_TEST_DIR, "files", "file", "base")
PROD_FILES = os.path.join(INTEGRATION_TEST_DIR, "files", "file", "prod")
PYEXEC = "python{}.{}".format(*sys.version_info)
MOCKBIN = os.path.join(INTEGRATION_TEST_DIR, "mockbin")
SCRIPT_DIR = os.path.join(CODE_DIR, "scripts")
TMP_STATE_TREE = os.path.join(SYS_TMP_DIR, "salt-temp-state-tree")
TMP_PILLAR_TREE = os.path.join(SYS_TMP_DIR, "salt-temp-pillar-tree")
TMP_PRODENV_STATE_TREE = os.path.join(SYS_TMP_DIR, "salt-temp-prodenv-state-tree")
TMP_PRODENV_PILLAR_TREE = os.path.join(SYS_TMP_DIR, "salt-temp-prodenv-pillar-tree")
TMP_CONF_DIR = TMP_MINION_CONF_DIR = os.path.join(TMP, "config")
TMP_SUB_MINION_CONF_DIR = os.path.join(TMP_CONF_DIR, "sub-minion")
TMP_SYNDIC_MINION_CONF_DIR = os.path.join(TMP_CONF_DIR, "syndic-minion")
TMP_SYNDIC_MASTER_CONF_DIR = os.path.join(TMP_CONF_DIR, "syndic-master")
TMP_SSH_CONF_DIR = TMP_MINION_CONF_DIR
CONF_DIR = os.path.join(INTEGRATION_TEST_DIR, "files", "conf")
PILLAR_DIR = os.path.join(FILES, "pillar")
TMP_SCRIPT_DIR = os.path.join(TMP, "scripts")
ENGINES_DIR = os.path.join(FILES, "engines")
LOG_HANDLERS_DIR = os.path.join(FILES, "log_handlers")


def list_test_mods():
    """
    A generator which returns all of the test files
    """
    test_re = re.compile(r"^test_.+\.py$")
    for dirname in (UNIT_TEST_DIR, INTEGRATION_TEST_DIR):
        test_type = os.path.basename(dirname)
        for root, _, files in os.walk(dirname):
            parent_mod = root[len(dirname) :].lstrip(os.sep).replace(os.sep, ".")
            for filename in files:
                if test_re.match(filename):
                    mod_name = test_type
                    if parent_mod:
                        mod_name += "." + parent_mod
                    mod_name += "." + filename[:-3]
                    yield mod_name



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>

    Test cases for salt.modules.locate
"""


import pytest

import salt.modules.locate as locate
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {locate: {}}


# 'version' function tests: 1


def test_version():
    """
    Test if it returns the version of locate
    """
    mock = MagicMock(return_value="mlocate 0.26")
    with patch.dict(locate.__salt__, {"cmd.run": mock}):
        assert locate.version() == ["mlocate 0.26"]


# 'stats' function tests: 1


def test_stats():
    """
    Test if it returns statistics about the locate database
    """
    ret = {
        "files": "75,253",
        "directories": "49,252",
        "bytes in file names": "93,214",
        "bytes used to store database": "29,165",
        "database": "/var/lib/mlocate/mlocate.db",
    }

    mock_ret = """Database /var/lib/mlocate/mlocate.db:
    49,252 directories
    75,253 files
    93,214 bytes in file names
    29,165 bytes used to store database"""

    with patch.dict(locate.__salt__, {"cmd.run": MagicMock(return_value=mock_ret)}):
        assert locate.stats() == ret


# 'updatedb' function tests: 1


def test_updatedb():
    """
    Test if it updates the locate database
    """
    mock = MagicMock(return_value="")
    with patch.dict(locate.__salt__, {"cmd.run": mock}):
        assert locate.updatedb() == []


# 'locate' function tests: 1


def test_locate():
    """
    Test if it performs a file lookup.
    """
    mock = MagicMock(return_value="")
    with patch.dict(locate.__salt__, {"cmd.run": mock}):
        assert locate.locate("wholename", database="myfile") == []



"""
Installing of certificates to the Windows Certificate Manager
=============================================================

Install certificates to the Windows Certificate Manager

.. code-block:: yaml

    salt://certs/cert.cer:
      certutil.add_store:
        - store: TrustedPublisher
"""

import logging

import salt.utils.platform

log = logging.getLogger(__name__)
__virtualname__ = "certutil"


def __virtual__():
    """
    Only work on Windows
    """
    if salt.utils.platform.is_windows():
        return __virtualname__
    return (False, "Only Windows supported")


def add_store(name, store, saltenv="base"):
    """
    Store a certificate to the given certificate store

    Args:

        name (str):
            The path to the certificate to add to the store. This is either the
            path to a local file or a file from the file server in the form of
            ``salt://path/to/file``

        store (str):
            The certificate store to add the certificate to

        saltenv (str):
            The salt environment to use. This is ignored if the path is local

    Returns:
        dict: A dictionary containing the results

    CLI Example:

    .. code-block:: yaml

        add_certificate:
          certutil.add_store:
            name: salt://web_cert.cer
            store: TrustedPublisher
    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    cert_file = __salt__["cp.cache_file"](name, saltenv)
    if cert_file is False:
        ret["comment"] = "Certificate file not found: {}".format(name)
        ret["result"] = False
        return ret

    cert_serial = __salt__["certutil.get_cert_serial"](name)
    if cert_serial is None:
        ret["comment"] = "Invalid certificate file: {}".format(name)
        ret["result"] = False
        return ret

    old_serials = __salt__["certutil.get_stored_cert_serials"](store=store)
    if cert_serial in old_serials:
        ret["comment"] = "Certificate already present: {}".format(name)
        return ret

    if __opts__["test"]:
        ret["comment"] = "Certificate will be added: {}".format(name)
        ret["result"] = None
        return ret

    retcode = __salt__["certutil.add_store"](name, store, retcode=True)
    if retcode != 0:
        ret["comment"] = "Error adding certificate: {}".format(name)
        ret["result"] = False
        return ret

    new_serials = __salt__["certutil.get_stored_cert_serials"](store=store)
    if cert_serial in new_serials:
        ret["changes"]["added"] = name
        ret["comment"] = "Added certificate: {}".format(name)
    else:
        ret["comment"] = "Failed to add certificate: {}".format(name)
        ret["result"] = False

    return ret


def del_store(name, store, saltenv="base"):
    """
    Remove a certificate from the given certificate store

    Args:

        name (str):
            The path to the certificate to remove from the store. This is either
            the path to a local file or a file from the file server in the form
            of ``salt://path/to/file``

        store (str):
            The certificate store to remove the certificate from

        saltenv (str):
            The salt environment to use. This is ignored if the path is local

    Returns:
        dict: A dictionary containing the results

    CLI Example:

    .. code-block:: yaml

        remove_certificate:
          certutil.del_store:
            name: salt://web_cert.cer
            store: TrustedPublisher
    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    cert_file = __salt__["cp.cache_file"](name, saltenv)
    if cert_file is False:
        ret["comment"] = "Certificate file not found: {}".format(name)
        ret["result"] = False
        return ret

    cert_serial = __salt__["certutil.get_cert_serial"](name)
    if cert_serial is None:
        ret["comment"] = "Invalid certificate file: {}".format(name)
        ret["result"] = False
        return ret

    old_serials = __salt__["certutil.get_stored_cert_serials"](store=store)
    if cert_serial not in old_serials:
        ret["comment"] = "Certificate already absent: {}".format(name)
        return ret

    if __opts__["test"]:
        ret["comment"] = "Certificate will be removed: {}".format(name)
        ret["result"] = None
        return ret

    retcode = __salt__["certutil.del_store"](name, store, retcode=True)
    if retcode != 0:
        ret["comment"] = "Error removing certificate: {}".format(name)
        ret["result"] = False
        return ret

    new_serials = __salt__["certutil.get_stored_cert_serials"](store=store)
    if cert_serial not in new_serials:
        ret["changes"]["removed"] = name
        ret["comment"] = "Removed certificate: {}".format(name)
    else:
        ret["comment"] = "Failed to remove certificate: {}".format(name)
        ret["result"] = False

    return ret



"""
    Test cases for salt.modules.gentoo_service
"""


import pytest

import salt.modules.gentoo_service as gentoo_service
from tests.support.mock import MagicMock, call, patch


@pytest.fixture
def configure_loader_modules():
    return {gentoo_service: {}}


def test_service_list_parser():
    """
    Test for parser of rc-status results
    """
    # no services is enabled
    mock = MagicMock(return_value="")
    with patch.dict(gentoo_service.__salt__, {"cmd.run": mock}):
        assert not gentoo_service.get_enabled()
    mock.assert_called_once_with("rc-update -v show")


def test_get_enabled_single_runlevel():
    """
    Test for Return a list of service that are enabled on boot
    """
    service_name = "name"
    runlevels = ["default"]
    mock = MagicMock(return_value=__services({service_name: runlevels}))
    with patch.dict(gentoo_service.__salt__, {"cmd.run": mock}):
        enabled_services = gentoo_service.get_enabled()
        assert service_name in enabled_services
        assert enabled_services[service_name] == runlevels


def test_get_enabled_filters_out_disabled_services():
    """
    Test for Return a list of service that are enabled on boot
    """
    service_name = "name"
    runlevels = ["default"]
    disabled_service = "disabled"
    service_list = __services({service_name: runlevels, disabled_service: []})

    mock = MagicMock(return_value=service_list)
    with patch.dict(gentoo_service.__salt__, {"cmd.run": mock}):
        enabled_services = gentoo_service.get_enabled()
        assert len(enabled_services) == 1
        assert service_name in enabled_services
        assert enabled_services[service_name] == runlevels


def test_get_enabled_with_multiple_runlevels():
    """
    Test for Return a list of service that are enabled on boot at more than one runlevel
    """
    service_name = "name"
    runlevels = ["non-default", "default"]
    mock = MagicMock(return_value=__services({service_name: runlevels}))
    with patch.dict(gentoo_service.__salt__, {"cmd.run": mock}):
        enabled_services = gentoo_service.get_enabled()
        assert service_name in enabled_services
        assert enabled_services[service_name][0] == runlevels[1]
        assert enabled_services[service_name][1] == runlevels[0]


def test_get_disabled():
    """
    Test for Return a list of service that are installed but disabled
    """
    disabled_service = "disabled"
    enabled_service = "enabled"
    service_list = __services({disabled_service: [], enabled_service: ["default"]})
    mock = MagicMock(return_value=service_list)
    with patch.dict(gentoo_service.__salt__, {"cmd.run": mock}):
        disabled_services = gentoo_service.get_disabled()
        assert disabled_services, 1
        assert disabled_service in disabled_services


def test_available():
    """
    Test for Returns ``True`` if the specified service is
    available, otherwise returns
    ``False``.
    """
    disabled_service = "disabled"
    enabled_service = "enabled"
    multilevel_service = "multilevel"
    missing_service = "missing"
    shutdown_service = "shutdown"
    service_list = __services(
        {
            disabled_service: [],
            enabled_service: ["default"],
            multilevel_service: ["default", "shutdown"],
            shutdown_service: ["shutdown"],
        }
    )
    mock = MagicMock(return_value=service_list)
    with patch.dict(gentoo_service.__salt__, {"cmd.run": mock}):
        assert gentoo_service.available(enabled_service)
        assert gentoo_service.available(multilevel_service)
        assert gentoo_service.available(disabled_service)
        assert gentoo_service.available(shutdown_service)
        assert not gentoo_service.available(missing_service)


def test_missing():
    """
    Test for The inverse of service.available.
    """
    disabled_service = "disabled"
    enabled_service = "enabled"
    service_list = __services({disabled_service: [], enabled_service: ["default"]})
    mock = MagicMock(return_value=service_list)
    with patch.dict(gentoo_service.__salt__, {"cmd.run": mock}):
        assert not gentoo_service.missing(enabled_service)
        assert not gentoo_service.missing(disabled_service)
        assert gentoo_service.missing("missing")


def test_getall():
    """
    Test for Return all available boot services
    """
    disabled_service = "disabled"
    enabled_service = "enabled"
    service_list = __services({disabled_service: [], enabled_service: ["default"]})
    mock = MagicMock(return_value=service_list)
    with patch.dict(gentoo_service.__salt__, {"cmd.run": mock}):
        all_services = gentoo_service.get_all()
        assert len(all_services) == 2
        assert disabled_service in all_services
        assert enabled_service in all_services


def test_start():
    """
    Test for Start the specified service
    """
    mock = MagicMock(return_value=True)
    with patch.dict(gentoo_service.__salt__, {"cmd.retcode": mock}):
        assert not gentoo_service.start("name")
    mock.assert_called_once_with(
        "/etc/init.d/name start", ignore_retcode=False, python_shell=False
    )


def test_stop():
    """
    Test for Stop the specified service
    """
    mock = MagicMock(return_value=True)
    with patch.dict(gentoo_service.__salt__, {"cmd.retcode": mock}):
        assert not gentoo_service.stop("name")
    mock.assert_called_once_with(
        "/etc/init.d/name stop", ignore_retcode=False, python_shell=False
    )


def test_restart():
    """
    Test for Restart the named service
    """
    mock = MagicMock(return_value=True)
    with patch.dict(gentoo_service.__salt__, {"cmd.retcode": mock}):
        assert not gentoo_service.restart("name")
    mock.assert_called_once_with(
        "/etc/init.d/name restart", ignore_retcode=False, python_shell=False
    )


def test_reload_():
    """
    Test for Reload the named service
    """
    mock = MagicMock(return_value=True)
    with patch.dict(gentoo_service.__salt__, {"cmd.retcode": mock}):
        assert not gentoo_service.reload_("name")
    mock.assert_called_once_with(
        "/etc/init.d/name reload", ignore_retcode=False, python_shell=False
    )


def test_zap():
    """
    Test for Reload the named service
    """
    mock = MagicMock(return_value=True)
    with patch.dict(gentoo_service.__salt__, {"cmd.retcode": mock}):
        assert not gentoo_service.zap("name")
    mock.assert_called_once_with(
        "/etc/init.d/name zap", ignore_retcode=False, python_shell=False
    )


def test_status():
    """
    Test for Return the status for a service
    """
    mock = MagicMock(return_value=True)
    with patch.dict(gentoo_service.__salt__, {"status.pid": mock}):
        assert gentoo_service.status("name", 1)

    # service is running
    mock = MagicMock(return_value=0)
    with patch.dict(gentoo_service.__salt__, {"cmd.retcode": mock}):
        assert gentoo_service.status("name")
    mock.assert_called_once_with(
        "/etc/init.d/name status", ignore_retcode=True, python_shell=False
    )

    # service is not running
    mock = MagicMock(return_value=1)
    with patch.dict(gentoo_service.__salt__, {"cmd.retcode": mock}):
        assert not gentoo_service.status("name")
    mock.assert_called_once_with(
        "/etc/init.d/name status", ignore_retcode=True, python_shell=False
    )

    # service is stopped
    mock = MagicMock(return_value=3)
    with patch.dict(gentoo_service.__salt__, {"cmd.retcode": mock}):
        assert not gentoo_service.status("name")
    mock.assert_called_once_with(
        "/etc/init.d/name status", ignore_retcode=True, python_shell=False
    )

    # service has crashed
    mock = MagicMock(return_value=32)
    with patch.dict(gentoo_service.__salt__, {"cmd.retcode": mock}):
        assert not gentoo_service.status("name")
    mock.assert_called_once_with(
        "/etc/init.d/name status", ignore_retcode=True, python_shell=False
    )


def test_enable():
    """
    Test for Enable the named service to start at boot
    """
    rc_update_mock = MagicMock(return_value=0)
    with patch.dict(gentoo_service.__salt__, {"cmd.retcode": rc_update_mock}):
        assert gentoo_service.enable("name")
    rc_update_mock.assert_called_once_with(
        "rc-update add name", ignore_retcode=False, python_shell=False
    )
    rc_update_mock.reset_mock()

    # move service from 'l1' to 'l2' runlevel
    service_name = "name"
    runlevels = ["l1"]
    level_list_mock = MagicMock(return_value=__services({service_name: runlevels}))
    with patch.dict(gentoo_service.__salt__, {"cmd.run": level_list_mock}):
        with patch.dict(gentoo_service.__salt__, {"cmd.retcode": rc_update_mock}):
            assert gentoo_service.enable("name", runlevels="l2")
    rc_update_mock.assert_has_calls(
        [
            call("rc-update delete name l1", ignore_retcode=False, python_shell=False),
            call("rc-update add name l2", ignore_retcode=False, python_shell=False),
        ]
    )
    rc_update_mock.reset_mock()

    # requested levels are the same as the current ones
    with patch.dict(gentoo_service.__salt__, {"cmd.run": level_list_mock}):
        with patch.dict(gentoo_service.__salt__, {"cmd.retcode": rc_update_mock}):
            assert gentoo_service.enable("name", runlevels="l1")
    assert rc_update_mock.call_count == 0
    rc_update_mock.reset_mock()

    # same as above with the list instead of the string
    with patch.dict(gentoo_service.__salt__, {"cmd.run": level_list_mock}):
        with patch.dict(gentoo_service.__salt__, {"cmd.retcode": rc_update_mock}):
            assert gentoo_service.enable("name", runlevels=["l1"])
    assert rc_update_mock.call_count == 0
    rc_update_mock.reset_mock()

    # add service to 'l2' runlevel
    with patch.dict(gentoo_service.__salt__, {"cmd.run": level_list_mock}):
        with patch.dict(gentoo_service.__salt__, {"cmd.retcode": rc_update_mock}):
            assert gentoo_service.enable("name", runlevels=["l2", "l1"])
    rc_update_mock.assert_called_once_with(
        "rc-update add name l2", ignore_retcode=False, python_shell=False
    )
    rc_update_mock.reset_mock()

    # remove service from 'l1' runlevel
    runlevels = ["l1", "l2"]
    level_list_mock = MagicMock(return_value=__services({service_name: runlevels}))
    with patch.dict(gentoo_service.__salt__, {"cmd.run": level_list_mock}):
        with patch.dict(gentoo_service.__salt__, {"cmd.retcode": rc_update_mock}):
            assert gentoo_service.enable("name", runlevels=["l2"])
    rc_update_mock.assert_called_once_with(
        "rc-update delete name l1", ignore_retcode=False, python_shell=False
    )
    rc_update_mock.reset_mock()

    # move service from 'l2' add to 'l3', leaving at l1
    with patch.dict(gentoo_service.__salt__, {"cmd.run": level_list_mock}):
        with patch.dict(gentoo_service.__salt__, {"cmd.retcode": rc_update_mock}):
            assert gentoo_service.enable("name", runlevels=["l1", "l3"])
    rc_update_mock.assert_has_calls(
        [
            call("rc-update delete name l2", ignore_retcode=False, python_shell=False),
            call("rc-update add name l3", ignore_retcode=False, python_shell=False),
        ]
    )
    rc_update_mock.reset_mock()

    # remove from l1, l3, and add to l2, l4, and leave at l5
    runlevels = ["l1", "l3", "l5"]
    level_list_mock = MagicMock(return_value=__services({service_name: runlevels}))
    with patch.dict(gentoo_service.__salt__, {"cmd.run": level_list_mock}):
        with patch.dict(gentoo_service.__salt__, {"cmd.retcode": rc_update_mock}):
            assert gentoo_service.enable("name", runlevels=["l2", "l4", "l5"])
    rc_update_mock.assert_has_calls(
        [
            call(
                "rc-update delete name l1 l3",
                ignore_retcode=False,
                python_shell=False,
            ),
            call("rc-update add name l2 l4", ignore_retcode=False, python_shell=False),
        ]
    )
    rc_update_mock.reset_mock()

    # rc-update failed
    rc_update_mock = MagicMock(return_value=1)
    with patch.dict(gentoo_service.__salt__, {"cmd.retcode": rc_update_mock}):
        assert not gentoo_service.enable("name")
    rc_update_mock.assert_called_once_with(
        "rc-update add name", ignore_retcode=False, python_shell=False
    )
    rc_update_mock.reset_mock()

    # move service delete failed
    runlevels = ["l1"]
    level_list_mock = MagicMock(return_value=__services({service_name: runlevels}))
    with patch.dict(gentoo_service.__salt__, {"cmd.run": level_list_mock}):
        with patch.dict(gentoo_service.__salt__, {"cmd.retcode": rc_update_mock}):
            assert not gentoo_service.enable("name", runlevels="l2")
    rc_update_mock.assert_called_once_with(
        "rc-update delete name l1", ignore_retcode=False, python_shell=False
    )
    rc_update_mock.reset_mock()

    # move service delete succeeds. add fails
    rc_update_mock = MagicMock()
    rc_update_mock.side_effect = [0, 1]
    with patch.dict(gentoo_service.__salt__, {"cmd.run": level_list_mock}):
        with patch.dict(gentoo_service.__salt__, {"cmd.retcode": rc_update_mock}):
            assert not gentoo_service.enable("name", runlevels="l2")
    rc_update_mock.assert_has_calls(
        [
            call("rc-update delete name l1", ignore_retcode=False, python_shell=False),
            call("rc-update add name l2", ignore_retcode=False, python_shell=False),
        ]
    )
    rc_update_mock.reset_mock()


def test_disable():
    """
    Test for Disable the named service to start at boot
    """
    rc_update_mock = MagicMock(return_value=0)
    with patch.dict(gentoo_service.__salt__, {"cmd.retcode": rc_update_mock}):
        assert gentoo_service.disable("name")
    rc_update_mock.assert_called_once_with(
        "rc-update delete name", ignore_retcode=False, python_shell=False
    )
    rc_update_mock.reset_mock()

    # disable service
    service_name = "name"
    runlevels = ["l1"]
    level_list_mock = MagicMock(return_value=__services({service_name: runlevels}))
    with patch.dict(gentoo_service.__salt__, {"cmd.run": level_list_mock}):
        with patch.dict(gentoo_service.__salt__, {"cmd.retcode": rc_update_mock}):
            assert gentoo_service.disable("name", runlevels="l1")
    rc_update_mock.assert_called_once_with(
        "rc-update delete name l1", ignore_retcode=False, python_shell=False
    )
    rc_update_mock.reset_mock()

    # same as above with list
    runlevels = ["l1"]
    level_list_mock = MagicMock(return_value=__services({service_name: runlevels}))
    with patch.dict(gentoo_service.__salt__, {"cmd.run": level_list_mock}):
        with patch.dict(gentoo_service.__salt__, {"cmd.retcode": rc_update_mock}):
            assert gentoo_service.disable("name", runlevels=["l1"])
    rc_update_mock.assert_called_once_with(
        "rc-update delete name l1", ignore_retcode=False, python_shell=False
    )
    rc_update_mock.reset_mock()

    # remove from 'l1', and leave at 'l2'
    runlevels = ["l1", "l2"]
    level_list_mock = MagicMock(return_value=__services({service_name: runlevels}))
    with patch.dict(gentoo_service.__salt__, {"cmd.run": level_list_mock}):
        with patch.dict(gentoo_service.__salt__, {"cmd.retcode": rc_update_mock}):
            assert gentoo_service.disable("name", runlevels=["l1"])
    rc_update_mock.assert_called_once_with(
        "rc-update delete name l1", ignore_retcode=False, python_shell=False
    )
    rc_update_mock.reset_mock()

    # remove from non-enabled level
    runlevels = ["l2"]
    level_list_mock = MagicMock(return_value=__services({service_name: runlevels}))
    with patch.dict(gentoo_service.__salt__, {"cmd.run": level_list_mock}):
        with patch.dict(gentoo_service.__salt__, {"cmd.retcode": rc_update_mock}):
            assert gentoo_service.disable("name", runlevels=["l1"])
    assert rc_update_mock.call_count == 0
    rc_update_mock.reset_mock()

    # remove from 'l1' and 'l3', leave at 'l2'
    runlevels = ["l1", "l2", "l3"]
    level_list_mock = MagicMock(return_value=__services({service_name: runlevels}))
    with patch.dict(gentoo_service.__salt__, {"cmd.run": level_list_mock}):
        with patch.dict(gentoo_service.__salt__, {"cmd.retcode": rc_update_mock}):
            assert gentoo_service.disable("name", runlevels=["l1", "l3"])
    rc_update_mock.assert_called_once_with(
        "rc-update delete name l1 l3", ignore_retcode=False, python_shell=False
    )
    rc_update_mock.reset_mock()

    # rc-update failed
    rc_update_mock = MagicMock(return_value=1)
    with patch.dict(gentoo_service.__salt__, {"cmd.retcode": rc_update_mock}):
        assert not gentoo_service.disable("name")
    rc_update_mock.assert_called_once_with(
        "rc-update delete name", ignore_retcode=False, python_shell=False
    )
    rc_update_mock.reset_mock()

    # move service delete failed
    runlevels = ["l1"]
    level_list_mock = MagicMock(return_value=__services({service_name: runlevels}))
    with patch.dict(gentoo_service.__salt__, {"cmd.run": level_list_mock}):
        with patch.dict(gentoo_service.__salt__, {"cmd.retcode": rc_update_mock}):
            assert not gentoo_service.disable("name", runlevels="l1")
    rc_update_mock.assert_called_once_with(
        "rc-update delete name l1", ignore_retcode=False, python_shell=False
    )
    rc_update_mock.reset_mock()

    # move service delete succeeds. add fails
    runlevels = ["l1", "l2", "l3"]
    level_list_mock = MagicMock(return_value=__services({service_name: runlevels}))
    with patch.dict(gentoo_service.__salt__, {"cmd.run": level_list_mock}):
        with patch.dict(gentoo_service.__salt__, {"cmd.retcode": rc_update_mock}):
            assert not gentoo_service.disable("name", runlevels=["l1", "l3"])
    rc_update_mock.assert_called_once_with(
        "rc-update delete name l1 l3", ignore_retcode=False, python_shell=False
    )
    rc_update_mock.reset_mock()


def test_enabled():
    """
    Test for Return True if the named service is enabled, false otherwise
    """
    mock = MagicMock(return_value={"name": ["default"]})
    with patch.object(gentoo_service, "get_enabled", mock):
        # service is enabled at any level
        assert gentoo_service.enabled("name")
        # service is enabled at the requested runlevels
        assert gentoo_service.enabled("name", runlevels="default")
        # service is enabled at a different runlevels
        assert not gentoo_service.enabled("name", runlevels="boot")

    mock = MagicMock(return_value={"name": ["boot", "default"]})
    with patch.object(gentoo_service, "get_enabled", mock):
        # service is enabled at any level
        assert gentoo_service.enabled("name")
        # service is enabled at the requested runlevels
        assert gentoo_service.enabled("name", runlevels="default")
        # service is enabled at all levels
        assert gentoo_service.enabled("name", runlevels=["boot", "default"])
        # service is enabled at a different runlevels
        assert not gentoo_service.enabled("name", runlevels="some-other-level")
        # service is enabled at a different runlevels
        assert not gentoo_service.enabled(
            "name", runlevels=["boot", "some-other-level"]
        )


def test_disabled():
    """
    Test for Return True if the named service is disabled, false otherwise
    """
    mock = MagicMock(return_value=["name"])
    with patch.object(gentoo_service, "get_disabled", mock):
        assert gentoo_service.disabled("name")


def __services(services):
    return "\n".join([" | ".join([svc, " ".join(services[svc])]) for svc in services])



"""
Base classes for napalm unit tests

:codeauthor: :email:`Anthony Shaw <anthonyshaw@apache.org>`
"""


from functools import wraps

from salt.utils.immutabletypes import freeze

TEST_INTERFACES = freeze(
    {
        "Management1": {
            "is_up": False,
            "is_enabled": False,
            "description": "",
            "last_flapped": -1,
            "speed": 1000,
            "mac_address": "dead:beef:dead",
        }
    }
)

# Test data
TEST_FACTS = freeze(
    {
        "__opts__": {},
        "OPTIONAL_ARGS": {},
        "uptime": "Forever",
        "UP": True,
        "HOSTNAME": "test-device.com",
        "hostname": "test-device.com",
        "username": "admin",
        "os_version": "1.2.3",
        "model": "test_model",
        "serial_number": "123456",
        "vendor": "cisco",
        "interface_list": TEST_INTERFACES,
    }
)

TEST_ENVIRONMENT = freeze({"hot": "yes"})

TEST_COMMAND_RESPONSE = freeze({"show run": "all the command output"})

TEST_TRACEROUTE_RESPONSE = freeze(
    {
        "success": {
            1: {
                "probes": {
                    1: {
                        "rtt": 1.123,
                        "ip_address": "206.223.116.21",
                        "host_name": "eqixsj-google-gige.google.com",
                    }
                }
            }
        }
    }
)

TEST_PING_RESPONSE = freeze(
    {
        "success": {
            "probes_sent": 5,
            "packet_loss": 0,
            "rtt_min": 72.158,
            "rtt_max": 72.433,
            "rtt_avg": 72.268,
            "rtt_stddev": 0.094,
            "results": [{"ip_address": "1.1.1.1", "rtt": 72.248}],
        }
    }
)

TEST_ARP_TABLE = freeze(
    [
        {
            "interface": "MgmtEth0/RSP0/CPU0/0",
            "mac": "5C:5E:AB:DA:3C:F0",
            "ip": "172.17.17.1",
            "age": 1454496274.84,
        }
    ]
)

TEST_IPADDRS = freeze(
    {"FastEthernet8": {"ipv4": {"10.66.43.169": {"prefix_length": 22}}}}
)

TEST_INTERFACES = freeze(
    {
        "Management1": {
            "is_up": False,
            "is_enabled": False,
            "description": "",
            "last_flapped": -1,
            "speed": 1000,
            "mac_address": "dead:beef:dead",
        }
    }
)

TEST_LLDP_NEIGHBORS = freeze(
    {"Ethernet2": [{"hostname": "junos-unittest", "port": "520"}]}
)

TEST_MAC_TABLE = freeze(
    [
        {
            "mac": "00:1C:58:29:4A:71",
            "interface": "Ethernet47",
            "vlan": 100,
            "static": False,
            "active": True,
            "moves": 1,
            "last_move": 1454417742.58,
        }
    ]
)

TEST_RUNNING_CONFIG = freeze({"one": "two"})

TEST_OPTICS = freeze(
    {
        "et1": {
            "physical_channels": {
                "channel": [
                    {
                        "index": 0,
                        "state": {
                            "input_power": {
                                "instant": 0.0,
                                "avg": 0.0,
                                "min": 0.0,
                                "max": 0.0,
                            },
                            "output_power": {
                                "instant": 0.0,
                                "avg": 0.0,
                                "min": 0.0,
                                "max": 0.0,
                            },
                            "laser_bias_current": {
                                "instant": 0.0,
                                "avg": 0.0,
                                "min": 0.0,
                                "max": 0.0,
                            },
                        },
                    }
                ]
            }
        }
    }
)

TEST_BGP_CONFIG = freeze({"test": "value"})

TEST_BGP_NEIGHBORS = freeze(
    {
        "default": {
            8121: [
                {
                    "up": True,
                    "local_as": 13335,
                    "remote_as": 8121,
                    "local_address": "172.101.76.1",
                    "local_address_configured": True,
                    "local_port": 179,
                    "remote_address": "192.247.78.0",
                    "router_id": "192.168.0.1",
                    "remote_port": 58380,
                    "multihop": False,
                    "import_policy": "4-NTT-TRANSIT-IN",
                    "export_policy": "4-NTT-TRANSIT-OUT",
                    "input_messages": 123,
                    "output_messages": 13,
                    "input_updates": 123,
                    "output_updates": 5,
                    "messages_queued_out": 23,
                    "connection_state": "Established",
                    "previous_connection_state": "EstabSync",
                    "last_event": "RecvKeepAlive",
                    "suppress_4byte_as": False,
                    "local_as_prepend": False,
                    "holdtime": 90,
                    "configured_holdtime": 90,
                    "keepalive": 30,
                    "configured_keepalive": 30,
                    "active_prefix_count": 132808,
                    "received_prefix_count": 566739,
                    "accepted_prefix_count": 566479,
                    "suppressed_prefix_count": 0,
                    "advertise_prefix_count": 0,
                    "flap_count": 27,
                }
            ]
        }
    }
)

TEST_TERM_CONFIG = freeze({"result": True, "already_configured": False})

TEST_NTP_PEERS = freeze(
    {
        "192.168.0.1": 1,
        "172.17.17.1": 2,
        "172.17.17.2": 3,
        "2400:cb00:6:1024::c71b:840a": 4,
    }
)

TEST_NTP_SERVERS = freeze(
    {
        "192.168.0.1": 1,
        "172.17.17.1": 2,
        "172.17.17.2": 3,
        "2400:cb00:6:1024::c71b:840a": 4,
    }
)

TEST_NTP_STATS = freeze(
    [
        {
            "remote": "188.114.101.4",
            "referenceid": "188.114.100.1",
            "synchronized": True,
            "stratum": 4,
            "type": "-",
            "when": "107",
            "hostpoll": 256,
            "reachability": 377,
            "delay": 164.228,
            "offset": -13.866,
            "jitter": 2.695,
        }
    ]
)

TEST_PROBES_CONFIG = freeze(
    {
        "probe1": {
            "test1": {
                "probe_type": "icmp-ping",
                "target": "192.168.0.1",
                "source": "192.168.0.2",
                "probe_count": 13,
                "test_interval": 3,
            },
            "test2": {
                "probe_type": "http-ping",
                "target": "172.17.17.1",
                "source": "192.17.17.2",
                "probe_count": 5,
                "test_interval": 60,
            },
        }
    }
)

TEST_PROBES_RESULTS = freeze(
    {
        "probe1": {
            "test1": {
                "last_test_min_delay": 63.120,
                "global_test_min_delay": 62.912,
                "current_test_avg_delay": 63.190,
                "global_test_max_delay": 177.349,
                "current_test_max_delay": 63.302,
                "global_test_avg_delay": 63.802,
                "last_test_avg_delay": 63.438,
                "last_test_max_delay": 65.356,
                "probe_type": "icmp-ping",
                "rtt": 63.138,
                "last_test_loss": 0,
                "round_trip_jitter": -59.0,
                "target": "192.168.0.1",
                "source": "192.168.0.2",
                "probe_count": 15,
                "current_test_min_delay": 63.138,
            },
            "test2": {
                "last_test_min_delay": 176.384,
                "global_test_min_delay": 169.226,
                "current_test_avg_delay": 177.098,
                "global_test_max_delay": 292.628,
                "current_test_max_delay": 180.055,
                "global_test_avg_delay": 177.959,
                "last_test_avg_delay": 177.178,
                "last_test_max_delay": 184.671,
                "probe_type": "icmp-ping",
                "rtt": 176.449,
                "last_test_loss": 0,
                "round_trip_jitter": -34.0,
                "target": "172.17.17.1",
                "source": "172.17.17.2",
                "probe_count": 15,
                "current_test_min_delay": 176.402,
            },
        }
    }
)

TEST_ROUTE = freeze(
    {
        "172.16.0.0/25": [
            {
                "protocol": "BGP",
                "last_active": True,
                "current_active": True,
                "age": 1178693,
                "routing_table": "inet.0",
                "next_hop": "192.168.0.11",
                "outgoing_interface": "xe-1/1/1.100",
                "preference": 170,
                "selected_next_hop": False,
                "protocol_attributes": {
                    "remote_as": 65001,
                    "metric": 5,
                    "local_as": 13335,
                    "as_path": "",
                    "remote_address": "192.168.0.11",
                    "metric2": 0,
                    "local_preference": 0,
                    "communities": ["0:2", "no-export"],
                    "preference2": -1,
                },
                "inactive_reason": "",
            },
            {
                "protocol": "BGP",
                "last_active": False,
                "current_active": False,
                "age": 2359429,
                "routing_table": "inet.0",
                "next_hop": "192.168.0.17",
                "outgoing_interface": "xe-1/1/1.100",
                "preference": 170,
                "selected_next_hop": True,
                "protocol_attributes": {
                    "remote_as": 65001,
                    "metric": 5,
                    "local_as": 13335,
                    "as_path": "",
                    "remote_address": "192.168.0.17",
                    "metric2": 0,
                    "local_preference": 0,
                    "communities": ["0:3", "no-export"],
                    "preference2": -1,
                },
                "inactive_reason": "Not Best in its group - Router ID",
            },
        ]
    }
)

TEST_SNMP_INFO = freeze({"test_": "value"})

TEST_USERS = freeze(
    {
        "mircea": {
            "level": 15,
            "password": "$1$0P70xKPa$4jt5/10cBTckk6I/w/",
            "sshkeys": [
                "ssh-rsa "
                "AAAAB3NzaC1yc2EAAAADAQABAAABAQC4pFn+shPwTb2yELO4L7NtQrKOJXNeCl1je"
                "l9STXVaGnRAnuc2PXl35vnWmcUq6YbUEcgUTRzzXfmelJKuVJTJIlMXii7h2xkbQp0YZIEs4P"
                "8ipwnRBAxFfk/ZcDsdfsdfsdfsdN56ejk345jhk345jk345jk341p3A/9LIL7l6YewLBCwJj6"
                "D+fWSJ0/YW+7oH17Fk2HH+tw0L5PcWLHkwA4t60iXn16qDbIk/ze6jv2hDGdCdz7oYQeCE55C"
                "CHOHMJWYfN3jcL4s0qv8/u6Ka1FVkV7iMmro7ChThoV/5snI4Ljf2wKqgHH7TfNaCfpU0WvHA"
                "nTs8zhOrGScSrtb mircea@master-roshi"
            ],
        }
    }
)


class MockNapalmDevice:
    """Setup a mock device for our tests"""

    def __init__(self):
        self._d = {}

    def __getitem__(self, key):
        return self._d[key]

    def __setitem__(self, key, value):
        self._d[key] = value

    def get_facts(self):
        return TEST_FACTS.copy()

    def get_environment(self):
        return TEST_ENVIRONMENT.copy()

    def get_arp_table(self):
        return TEST_ARP_TABLE.copy()

    def get(self, key, default=None, *args, **kwargs):
        try:
            if key == "DRIVER":
                return self
            return TEST_FACTS.copy()[key]
        except KeyError:
            return default

    def cli(self, commands, *args, **kwargs):
        assert commands[0] == "show run"
        return TEST_COMMAND_RESPONSE.copy()

    def traceroute(self, destination, **kwargs):
        assert destination == "destination.com"
        return TEST_TRACEROUTE_RESPONSE.copy()

    def ping(self, destination, **kwargs):
        assert destination == "destination.com"
        return TEST_PING_RESPONSE.copy()

    def get_config(self, retrieve="all"):
        assert retrieve == "running"
        return TEST_RUNNING_CONFIG.copy()

    def get_interfaces_ip(self, **kwargs):
        return TEST_IPADDRS.copy()

    def get_interfaces(self, **kwargs):
        return TEST_INTERFACES.copy()

    def get_lldp_neighbors_detail(self, **kwargs):
        return TEST_LLDP_NEIGHBORS.copy()

    def get_mac_address_table(self, **kwargs):
        return TEST_MAC_TABLE.copy()

    def get_optics(self, **kwargs):
        return TEST_OPTICS.copy()

    def load_merge_candidate(self, filename=None, config=None):
        assert config == "new config"
        return TEST_RUNNING_CONFIG.copy()

    def load_replace_candidate(self, filename=None, config=None):
        assert config == "new config"
        return TEST_RUNNING_CONFIG.copy()

    def commit_config(self, **kwargs):
        return TEST_RUNNING_CONFIG.copy()

    def discard_config(self, **kwargs):
        return TEST_RUNNING_CONFIG.copy()

    def compare_config(self, **kwargs):
        return TEST_RUNNING_CONFIG.copy()

    def rollback(self, **kwargs):
        return TEST_RUNNING_CONFIG.copy()

    def get_bgp_config(self, **kwargs):
        return TEST_BGP_CONFIG.copy()

    def get_bgp_neighbors_detail(self, neighbor_address=None, **kwargs):
        assert neighbor_address is None or "test_address"
        return TEST_BGP_NEIGHBORS.copy()

    def get_ntp_peers(self, **kwargs):
        return TEST_NTP_PEERS.copy()

    def get_ntp_servers(self, **kwargs):
        return TEST_NTP_SERVERS.copy()

    def get_ntp_stats(self, **kwargs):
        return TEST_NTP_STATS.copy()

    def get_probes_config(self, **kwargs):
        return TEST_PROBES_CONFIG.copy()

    def get_probes_results(self, **kwargs):
        return TEST_PROBES_RESULTS.copy()

    def get_route_to(self, destination, protocol=None, **kwargs):
        assert destination == "1.2.3.4"
        return TEST_ROUTE.copy()

    def get_snmp_information(self, **kwargs):
        return TEST_SNMP_INFO.copy()

    def get_users(self, **kwargs):
        return TEST_USERS.copy()


def mock_proxy_napalm_wrap(func):
    """
    The proper decorator checks for proxy minions. We don't care
    so just pass back to the origination function
    """

    @wraps(func)
    def func_wrapper(*args, **kwargs):
        func.__globals__["napalm_device"] = MockNapalmDevice()
        return func(*args, **kwargs)

    return func_wrapper


def true(name):
    return True


def random_hash(source, method):
    return 12346789


def join(*files):
    return True


def get_managed_file(*args, **kwargs):
    return "True"



"""
znc - An advanced IRC bouncer

.. versionadded:: 2014.7.0

Provides an interface to basic ZNC functionality
"""

import hashlib
import logging
import os.path
import random
import signal

import salt.utils.path

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only load the module if znc is installed
    """
    if salt.utils.path.which("znc"):
        return "znc"
    return (False, "Module znc: znc binary not found")


def _makepass(password, hasher="sha256"):
    """
    Create a znc compatible hashed password
    """
    # Setup the hasher
    if hasher == "sha256":
        h = hashlib.sha256(password)
    elif hasher == "md5":
        h = hashlib.md5(password)
    else:
        return NotImplemented

    c = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789!?.,:;/*-+_()"
    r = {
        "Method": h.name,
        "Salt": "".join(random.SystemRandom().choice(c) for x in range(20)),
    }

    # Salt the password hash
    h.update(r["Salt"])
    r["Hash"] = h.hexdigest()

    return r


def buildmod(*modules):
    """
    Build module using znc-buildmod

    CLI Example:

    .. code-block:: bash

        salt '*' znc.buildmod module.cpp [...]
    """
    # Check if module files are missing
    missing = [module for module in modules if not os.path.exists(module)]
    if missing:
        return "Error: The file ({}) does not exist.".format(", ".join(missing))

    cmd = ["znc-buildmod"]
    cmd.extend(modules)
    out = __salt__["cmd.run"](cmd, python_shell=False).splitlines()
    return out[-1]


def dumpconf():
    """
    Write the active configuration state to config file

    CLI Example:

    .. code-block:: bash

        salt '*' znc.dumpconf
    """
    return __salt__["ps.pkill"]("znc", signal=signal.SIGUSR1)


def rehashconf():
    """
    Rehash the active configuration state from config file

    CLI Example:

    .. code-block:: bash

        salt '*' znc.rehashconf
    """
    return __salt__["ps.pkill"]("znc", signal=signal.SIGHUP)


def version():
    """
    Return server version from znc --version

    CLI Example:

    .. code-block:: bash

        salt '*' znc.version
    """
    cmd = ["znc", "--version"]
    out = __salt__["cmd.run"](cmd, python_shell=False).splitlines()
    ret = out[0].split(" - ")
    return ret[0]



"""
    :codeauthor: Rupesh Tare <rupesht@saltstack.com>

    Test cases for salt.modules.hadoop
"""


import pytest

import salt.modules.hadoop as hadoop
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {hadoop: {}}


def test_version():
    """
    Test for Return version from hadoop version
    """
    mock = MagicMock(return_value="A \nB \n")
    with patch.dict(hadoop.__salt__, {"cmd.run": mock}):
        assert hadoop.version() == "B"


def test_dfs():
    """
    Test for Execute a command on DFS
    """
    with patch.object(hadoop, "_hadoop_cmd", return_value="A"):
        assert hadoop.dfs("command") == "A"

    assert hadoop.dfs() == "Error: command must be provided"


def test_dfs_present():
    """
    Test for Check if a file or directory is present on the distributed FS.
    """
    with patch.object(
        hadoop, "_hadoop_cmd", side_effect=["No such file or directory", "A"]
    ):
        assert not hadoop.dfs_present("path")
        assert hadoop.dfs_present("path")


def test_dfs_absent():
    """
    Test for Check if a file or directory is absent on the distributed FS.
    """
    with patch.object(
        hadoop, "_hadoop_cmd", side_effect=["No such file or directory", "A"]
    ):
        assert hadoop.dfs_absent("path")
        assert not hadoop.dfs_absent("path")


def test_namenode_format():
    """
    Test for Format a name node
    """
    with patch.object(hadoop, "_hadoop_cmd", return_value="A"):
        assert hadoop.namenode_format("force") == "A"



"""
Module for working with the Zenoss API

.. versionadded:: 2016.3.0

:configuration: This module requires a 'zenoss' entry in the master/minion config.

    For example:

    .. code-block:: yaml

        zenoss:
          hostname: https://zenoss.example.com
          username: admin
          password: admin123
          verify_ssl: True
          ca_bundle: /etc/ssl/certs/ca-certificates.crt
"""

import logging
import re

import salt.utils.http
import salt.utils.json

log = logging.getLogger(__name__)

__virtualname__ = "zenoss"


def __virtual__():
    """
    Only load if requests is installed
    """
    return __virtualname__


ROUTERS = {
    "MessagingRouter": "messaging",
    "EventsRouter": "evconsole",
    "ProcessRouter": "process",
    "ServiceRouter": "service",
    "DeviceRouter": "device",
    "NetworkRouter": "network",
    "TemplateRouter": "template",
    "DetailNavRouter": "detailnav",
    "ReportRouter": "report",
    "MibRouter": "mib",
    "ZenPackRouter": "zenpack",
}


def _session():
    """
    Create a session to be used when connecting to Zenoss.
    """

    config = __salt__["config.option"]("zenoss")
    return salt.utils.http.session(
        user=config.get("username"),
        password=config.get("password"),
        verify_ssl=config.get("verify_ssl", True),
        ca_bundle=config.get("ca_bundle"),
        headers={"Content-type": "application/json; charset=utf-8"},
    )


def _router_request(router, method, data=None):
    """
    Make a request to the Zenoss API router
    """
    if router not in ROUTERS:
        return False

    req_data = salt.utils.json.dumps(
        [dict(action=router, method=method, data=data, type="rpc", tid=1)]
    )

    config = __salt__["config.option"]("zenoss")
    log.debug("Making request to router %s with method %s", router, method)
    url = "{}/zport/dmd/{}_router".format(config.get("hostname"), ROUTERS[router])
    response = _session().post(url, data=req_data)

    # The API returns a 200 response code even whe auth is bad.
    # With bad auth, the login page is displayed. Here I search for
    # an element on the login form to determine if auth failed.
    if re.search('name="__ac_name"', response.content):
        log.error("Request failed. Bad username/password.")
        raise Exception("Request failed. Bad username/password.")

    return salt.utils.json.loads(response.content).get("result", None)


def _determine_device_class():
    """
    If no device class is given when adding a device, this helps determine
    """
    if __salt__["grains.get"]("kernel") == "Linux":
        return "/Server/Linux"


def find_device(device=None):
    """
    Find a device in Zenoss. If device not found, returns None.

    Parameters:
        device:         (Optional) Will use the grain 'fqdn' by default

    CLI Example:

    .. code-block:: bash

        salt '*' zenoss.find_device
    """

    data = [{"uid": "/zport/dmd/Devices", "params": {}, "limit": None}]
    all_devices = _router_request("DeviceRouter", "getDevices", data=data)
    for dev in all_devices["devices"]:
        if dev["name"] == device:
            # We need to save the has for later operations
            dev["hash"] = all_devices["hash"]
            log.info("Found device %s in Zenoss", device)
            return dev

    log.info("Unable to find device %s in Zenoss", device)
    return None


def device_exists(device=None):
    """
    Check to see if a device already exists in Zenoss.

    Parameters:
        device:         (Optional) Will use the grain 'fqdn' by default

    CLI Example:

    .. code-block:: bash

        salt '*' zenoss.device_exists
    """

    if not device:
        device = __salt__["grains.get"]("fqdn")

    if find_device(device):
        return True
    return False


def add_device(device=None, device_class=None, collector="localhost", prod_state=1000):
    """
    A function to connect to a zenoss server and add a new device entry.

    Parameters:
        device:         (Optional) Will use the grain 'fqdn' by default.
        device_class:   (Optional) The device class to use. If none, will determine based on kernel grain.
        collector:      (Optional) The collector to use for this device. Defaults to 'localhost'.
        prod_state:     (Optional) The prodState to set on the device. If none, defaults to 1000 ( production )

    CLI Example:

    .. code-block:: bash

        salt '*' zenoss.add_device
    """

    if not device:
        device = __salt__["grains.get"]("fqdn")

    if not device_class:
        device_class = _determine_device_class()
    log.info("Adding device %s to zenoss", device)
    data = dict(
        deviceName=device,
        deviceClass=device_class,
        model=True,
        collector=collector,
        productionState=prod_state,
    )
    response = _router_request("DeviceRouter", "addDevice", data=[data])
    return response


def set_prod_state(prod_state, device=None):
    """
    A function to set the prod_state in zenoss.

    Parameters:
        prod_state:     (Required) Integer value of the state
        device:         (Optional) Will use the grain 'fqdn' by default.

    CLI Example:

    .. code-block:: bash

        salt zenoss.set_prod_state 1000 hostname
    """

    if not device:
        device = __salt__["grains.get"]("fqdn")

    device_object = find_device(device)

    if not device_object:
        return "Unable to find a device in Zenoss for {}".format(device)

    log.info("Setting prodState to %d on %s device", prod_state, device)
    data = dict(
        uids=[device_object["uid"]],
        prodState=prod_state,
        hashcheck=device_object["hash"],
    )
    return _router_request("DeviceRouter", "setProductionState", [data])
